#include <ndis.h>
#include "gmm.h"

NDIS_STATUS
gmm_allocate_recv_item (NDIS_HANDLE handle, gmm_recv_item * item,
			NDIS_HANDLE packet_pool, NDIS_HANDLE buffer_pool)
{
  NDIS_STATUS status;

  NdisMAllocateSharedMemory (handle, GMM_MTU, FALSE, &item->va, &item->pa);
  if (item->va == NULL)
    {
      status = NDIS_STATUS_FAILURE;
      goto abort_with_nothing;
    }
  NdisAllocatePacket (&status, &item->packet, packet_pool);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_shmem;
    }
  NDIS_SET_PACKET_HEADER_SIZE (item->packet, GMM_ETHERNET_HEADER_SIZE);
  NdisAllocateBuffer (&status, &item->buffer, buffer_pool,
		      item->va + 2, GMM_MTU - 2);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_packet;
    }
  NdisChainBufferAtFront (item->packet, item->buffer);
  *(gmm_recv_item **) & item->packet->MiniportReserved = item;
  return NDIS_STATUS_SUCCESS;

 abort_with_packet:
  NdisFreePacket (item->packet);
 abort_with_shmem:
  NdisMFreeSharedMemory (handle, GMM_MTU, FALSE, item->va, item->pa);
 abort_with_nothing:
  return status;
}

NDIS_STATUS
gmm_free_recv_item (NDIS_HANDLE handle, gmm_recv_item * item,
		    NDIS_HANDLE packet_pool, NDIS_HANDLE buffer_pool)
{
  NdisFreeBuffer (item->buffer);
  NdisFreePacket (item->packet);
  NdisMFreeSharedMemory (handle, GMM_MTU, FALSE, item->va, item->pa);
}

NDIS_STATUS
gmm_allocate_recv_ring (gmm_Context * context)
{
  int i;
  int j;
  gmm_recv_item *item;
  NDIS_STATUS status;

  NdisAllocatePacketPool (&status, &context->recv_packet_pool,
			  GMM_NUM_RECV_BUFFERS, 4 * sizeof (PVOID));
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_nothing;
    }
  NdisAllocateBufferPool (&status, &context->recv_buffer_pool,
			  GMM_NUM_RECV_BUFFERS);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_packet_pool;
    }
  NdisAllocateSpinLock (&context->recv_lock);

  item = context->recv_item;
  for (i = 0; i < GMM_NUM_RECV_BUFFERS; ++i)
    {
      status = gmm_allocate_recv_item (context->miniportAdapterHandle, item,
				       context->recv_packet_pool,
				       context->recv_buffer_pool);
      if (status != NDIS_STATUS_SUCCESS)
	{
	  goto abort_with_items;
	}
      context->recv_ring[i] = item;
      ++item;
    }
  return NDIS_STATUS_SUCCESS;

 abort_with_items:
  item = context->recv_item;
  for (j = 0; j < i; ++j)
    {
      gmm_free_recv_item (context->miniportAdapterHandle, item,
			  context->recv_packet_pool,
			  context->recv_buffer_pool);
      ++item;
    }
  NdisFreeSpinLock (&context->recv_lock);
 abort_with_buffer_pool:
  NdisFreeBufferPool (context->recv_buffer_pool);
 abort_with_packet_pool:
  NdisFreePacketPool (context->recv_packet_pool);
 abort_with_nothing:
  return NDIS_STATUS_FAILURE;
}

void
gmm_provide_recv_ring (gmm_Context * context)
{
  int i;
  gmm_recv_item *item;
  gm_ethernet_segment_descriptor_t segment;

  item = context->recv_item;
  for (i = 0; i < GMM_NUM_RECV_BUFFERS; ++i)
    {
#if (GM_SIZEOF_DP_T == 4)
      segment.ptr = gm_hton_dp (NdisGetPhysicalAddressLow (item->pa));
#elif (GM_SIZEOF_DP_T == 8)
      segment.ptr = gm_hton_dp (item->pa.QuadPart);
#endif
      segment.len = gm_htonl (GMM_MTU);
      _gm_ethernet_provide_scatter_list (context->port, 1, &segment);
      ++item;
    }
}

void
gmm_free_recv_ring (gmm_Context * context)
{
  int i;
  gmm_recv_item *item;

  item = context->recv_item;
  for (i = 0; i < GMM_NUM_RECV_BUFFERS; ++i)
    {
      gmm_free_recv_item (context->miniportAdapterHandle, item,
			  context->recv_packet_pool,
			  context->recv_buffer_pool);
      ++item;
    }

  NdisFreeSpinLock (&context->recv_lock);
  NdisFreeBufferPool (context->recv_buffer_pool);
  NdisFreePacketPool (context->recv_packet_pool);
}

VOID
gmm_ReturnPacket (NDIS_HANDLE miniportAdapterContext, PNDIS_PACKET packet)
{
  gmm_Context *context = (gmm_Context *) miniportAdapterContext;
  gmm_recv_item *item;
  gm_ethernet_segment_descriptor_t segment;
  int head;
  int count;

  NdisAcquireSpinLock (&context->recv_lock);
  item = *((gmm_recv_item **) packet->MiniportReserved);
  head = context->recv_ring_head;
  count = context->recv_ring_count;
  context->recv_ring[((head + GMM_NUM_RECV_BUFFERS - count)
		      % GMM_NUM_RECV_BUFFERS)] = item;
  context->recv_ring_count--;

#if (GM_SIZEOF_DP_T == 4)
  segment.ptr = gm_hton_dp (NdisGetPhysicalAddressLow (item->pa));
#elif (GM_SIZEOF_DP_T == 8)
  segment.ptr = gm_hton_dp (item->pa.QuadPart);
#endif
  segment.len = gm_htonl (GMM_MTU);
  _gm_ethernet_provide_scatter_list (context->port, 1, &segment);
  NdisReleaseSpinLock (&context->recv_lock);
}

void
gmm_receiveInterrupt (gmm_Context * context, unsigned length, gm_u16_t csum)
{
  gmm_recv_item *item;

  NdisAcquireSpinLock (&context->recv_lock);
  item = context->recv_ring[context->recv_ring_head];
  context->recv_ring_head = ((context->recv_ring_head + 1)
			     % GMM_NUM_RECV_BUFFERS);
  context->recv_ring_count++;

  NdisAdjustBufferLength (item->buffer, length - GMM_HEADER_SIZE);
  NDIS_SET_PACKET_STATUS (item->packet, NDIS_STATUS_RESOURCES);
  context->rcv_ok++;
  NdisReleaseSpinLock (&context->recv_lock);
  NdisMIndicateReceivePacket (context->miniportAdapterHandle, &item->packet,
			      1);
  gmm_ReturnPacket (context, item->packet);
}
