/*************************************************************************
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *									 *
 * Portions Copyright 1999 by Cam Macdonnell - University of Alberta.	 *
 * Used with permission.						 *
 *************************************************************************/

/*
 * Simple Send/Recv for GM
 * Author: Max Stern (Myricom)
 * Based on a program by Cam Macdonell - University of Alberta
 *	
 *
 * Syntax: gm_simple_example_recv <sender_nodename> [<local_board_num>]
 *
 *         <sender_nodename> is required.  Include the board_num suffix if the
 *         sender is using a board_num other than 0; e.g., hosta:1 for board 1
 *         on host "hosta".
 *
 *         The default for <local_board_num> is 0.
 *
 * Description:
 *
 *   This program, together with gm_simple_example_send, serves as a super-
 *   simple example of sending and receiving in GM, using both the basic send
 *   (gm_send type) and the directed send (gm_directed_send type) paradigms.
 *
 *   This program, the "receiver" or slave process, takes the following main
 *   actions (see commentary in the code below for the "how" details):
 *
 *   - uses gm_send_to_peer_with_callback() to transmit its node_id and its
 *     directed-receive buffer address to the "sender" process (which may be
 *     on the same or a different host);
 *
 *   - loops on gm_receive() until both the send callback from the above send
 *     and an incoming message from gm_simple_example_send have been received
 *     (the incoming message serves as notification that the directed-receive
 *     buffer has been modified);
 *
 *   - displays the contents of its directed-receive buffer, showing the
 *     contents stored there by gm_simple_example_send;
 *
 *   - exits.
 *
 *   See also the Description and the Usage Note in gm_simple_example_send.c.
 *
 * */


#include <stdlib.h>
#include <stdio.h>
#include "gm.h"
#include "gm_simple_example.h"

/* A structure used to store the state of this simple example program. */

typedef struct
{
  int messages_expected;
  int callbacks_pending;
} gm_se_context_t;

static void wait_for_events (struct gm_port *p,
			     gm_se_context_t *the_context);
static void my_send_callback (struct gm_port *port,
			      void *the_context,
			      gm_status_t the_status);

int
main (int argc, char **argv)
{
  char sender_nodename[64];
  gm_se_context_t context;
  gm_status_t main_status;
  int my_board_num = 0;		/* Default board_num is 0 */
  struct gm_port *my_port;
  unsigned int my_node_id;
  unsigned int sender_node_id;
  void *directed_recv_buffer;
  void *in_buffer;
  void *out_buffer;

  context.messages_expected = 0;
  context.callbacks_pending = 0;

  gm_init();

  /* Parse the program argument(s) */
  
  if ((argc < 2) || (argc > 3))
    {
      gm_printf ("USAGE: gm_simple_example_recv <sender_nodename> "
		 "[<local_board_num>]\n");
      main_status = GM_INVALID_PARAMETER;
      goto abort_with_nothing;
    }
  if (gm_strlen (argv[1]) + 1 > sizeof (sender_nodename))
    {
      gm_printf ("[recv] *** ERROR: "
		 "sender nodename length %ld exceeds maximum of %ld\n",
		 gm_strlen (argv[1]), sizeof (sender_nodename) - 1);
      main_status = GM_INVALID_PARAMETER;
      goto abort_with_nothing;
    }
  gm_strncpy (sender_nodename,	/* Mandatory 1st parameter */
	      argv[1],
	      sizeof (sender_nodename) - 1);
  if (argc > 2)
    {
      my_board_num = atoi (argv[2]);	/* Optional 2nd parameter  */
    }

  /* Open a port on our local interface. */
  
  main_status = gm_open (&my_port, my_board_num, GM_SIMPLE_EXAMPLE_PORT_NUM,
			 "gm_simple_example_recv",
			 (enum gm_api_version) GM_API_VERSION_1_1);
  if (main_status == GM_SUCCESS)
    {
      gm_printf ("[recv] Opened board %d port %d\n",
		 my_board_num, GM_SIMPLE_EXAMPLE_PORT_NUM);
    }
  else
    {
      gm_perror ("[recv] Couldn't open GM port", main_status);
      goto abort_with_nothing;
    }

  /* Try to convert the sender's hostname into a node ID. */
  
  sender_node_id = gm_host_name_to_node_id (my_port, sender_nodename);
  if (sender_node_id == GM_NO_SUCH_NODE_ID)
    {
      gm_printf ("[recv] Nodename %s unknown\n", sender_nodename);
      main_status = GM_INVALID_PARAMETER;
      goto abort_with_open_port;
    }

  gm_get_node_id (my_port, &my_node_id);
  if (my_node_id == sender_node_id)
    {
      gm_printf ("[recv] *** ERROR: "
		 "sender and receiver are same node, id=%d\n",
		 my_node_id);
      main_status = GM_INVALID_PARAMETER;
      goto abort_with_open_port;
    }

  /* Allocate DMAable message buffers. */

  out_buffer = gm_dma_calloc (my_port, GM_SIMPLE_EXAMPLE_BUFFER_COUNT,
			      GM_SIMPLE_EXAMPLE_BUFFER_LENGTH);
  if (out_buffer == 0)
    {
      gm_printf ("[recv] Couldn't allocate out_buffer\n");
      main_status = GM_OUT_OF_MEMORY;
      goto abort_with_open_port;
    }

  in_buffer = gm_dma_calloc (my_port, GM_SIMPLE_EXAMPLE_BUFFER_COUNT,
			     GM_SIMPLE_EXAMPLE_BUFFER_LENGTH);
  if (in_buffer == 0)
    {
      gm_printf ("[recv] Couldn't allocate in_buffer\n");
      main_status = GM_OUT_OF_MEMORY;
      goto abort_with_out_buffer;
    }

  directed_recv_buffer = gm_dma_calloc (my_port,
					GM_SIMPLE_EXAMPLE_BUFFER_COUNT,
					GM_SIMPLE_EXAMPLE_BUFFER_LENGTH);
  if (directed_recv_buffer == 0)
    {
      gm_printf ("[recv] Couldn't allocate directed_recv_buffer\n");
      main_status = GM_OUT_OF_MEMORY;
      goto abort_with_in_buffer;
    }

  /* Allow any GM process to modify any of the local DMAable buffers. */

  gm_allow_remote_memory_access (my_port);

  /* Compose and send the messages to tell sender our node_id and our
     directed-receive-buffer address */

  sprintf (out_buffer, "%d", my_node_id);
  gm_send_to_peer_with_callback (my_port,
				 out_buffer,
				 GM_SIMPLE_EXAMPLE_SIZE,
				 (unsigned long) gm_strlen (out_buffer) + 1,
				 GM_SIMPLE_EXAMPLE_PRIORITY,
				 sender_node_id,
				 my_send_callback,
				 &context);
  context.callbacks_pending++;
  wait_for_events (my_port, &context); /* For simplicity, wait for callback
					  before going on */

  sprintf (out_buffer, "%p", directed_recv_buffer);
  gm_send_to_peer_with_callback (my_port,
				 out_buffer,
				 GM_SIMPLE_EXAMPLE_SIZE,
				 (unsigned long) gm_strlen (out_buffer) + 1,
				 GM_SIMPLE_EXAMPLE_PRIORITY,
				 sender_node_id,
				 my_send_callback,
				 &context);
  context.callbacks_pending++;

  gm_provide_receive_buffer (my_port, in_buffer, GM_SIMPLE_EXAMPLE_SIZE,
			     GM_SIMPLE_EXAMPLE_PRIORITY);

  /* Prefill the directed-receive buffer with garbage; we'll see it change */
  sprintf (directed_recv_buffer, "lots a filler, don't need this");

  context.messages_expected = 1; /* Now, we do expect a message from sender */

  /* Nothing more to do but wait for callbacks and the incoming messages */

  wait_for_events (my_port, &context);

  gm_printf ("[recv] "
	     "Having received the incoming message from the sender, the\n"
	     "       directed-receive buffer contains \"%s\"\n",
	     directed_recv_buffer);

  gm_printf ("[recv] gm_simple_example_recv completed successfully\n");

  gm_dma_free (my_port, in_buffer);
  gm_dma_free (my_port, out_buffer);
  gm_dma_free (my_port, directed_recv_buffer);
  gm_close (my_port);
  gm_exit (GM_SUCCESS);

 abort_with_in_buffer:
  gm_dma_free (my_port, in_buffer);
 abort_with_out_buffer:
  gm_dma_free (my_port, out_buffer);
 abort_with_open_port:
  gm_close (my_port);
 abort_with_nothing:
  gm_finalize();
  gm_exit (main_status);
}

static void
wait_for_events (struct gm_port *my_port, gm_se_context_t *the_context)
{
  gm_recv_event_t *event;
  void *buffer;
  unsigned int size;


  while ((the_context->callbacks_pending > 0)
	 || (the_context->messages_expected > 0))
    {
      event = gm_receive (my_port);

      switch (gm_ntoh_u8 (event->recv.type))
	{
	case GM_RECV_EVENT:
	case GM_PEER_RECV_EVENT:
	case GM_FAST_PEER_RECV_EVENT:
	  gm_printf ("[recv] Received incoming message: \"%s\"\n",
		     (char *) gm_ntohp (event->recv.message));
	  the_context->messages_expected--;
	  /* Return the buffer for reuse */
	  buffer = gm_ntohp (event->recv.buffer);
	  size = gm_ntoh_u8 (event->recv.size);
	  gm_provide_receive_buffer (my_port, buffer, size,
				     GM_SIMPLE_EXAMPLE_PRIORITY);
	  break;

	case GM_NO_RECV_EVENT:
	  break;

	default:
	  gm_unknown (my_port, event);
	}
    }
}

/* This function is called inside gm_unknown() when there is a callback
   ready to be processed.  It tells us that a send has completed, either
   successfully or with error. */

static void
my_send_callback (struct gm_port *port, void *the_context,
		  gm_status_t the_status)
{
  /* One pending callback has been received */
  ((gm_se_context_t *)the_context)->callbacks_pending--;

  if (the_status != GM_SUCCESS)
    {
      if (the_status != GM_SEND_DROPPED)
	{
	  gm_perror ("send completed with error", the_status);
	}
    }
}
