/*******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: glenn@myri.com */

/* initialize and finalize GM. */

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_enable_security.h"
#include "gm_internal.h"

/* Set this flag to 1 for gm_init()/gm_finalize() debugging */
#define GM_DEBUG_GM_INIT 0

#if GM_DEBUG_GM_INIT
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

#if GM_KERNEL
int _gm_initialized = 0;
struct gm_mutex *_gm_global_mutex = (struct gm_mutex *)0;
#else
int _gm_initialized;
struct gm_mutex *_gm_global_mutex;
#endif

#if defined WIN32 && !GM_KERNEL && GM_BUILDING_GM_LIB
#include <winsock.h>
static WSADATA gm_wsadata;
#endif

#ifdef DELETE_VXWORKS
#if GM_OS_VXWORKS
#include <setjmp.h>
extern jmp_buf gm_vxworks_fake_exit_env;
#endif
#endif

gm_status_t
gm_init ()
{
  gm_status_t status;
#if GM_PRINT_VERSION
  static int printed;
#endif
  unsigned long initialized;

  /* check for redundant inits. */

  if (_gm_initialized)
    {
      GM_PRINT(GM_DEBUG_GM_INIT,
               ("gm_init() bypassed; already initialized (%d)\n",
		_gm_initialized));
      goto done;
    }

  GM_PRINT(GM_DEBUG_GM_INIT, ("gm_init() entered for first time\n"));

#if GM_PRINT_VERSION
  if (!printed)
    {
      GM_INFO (("\n"));
      _GM_INFO (("Using GM version \"%s\" build \"%s.\"\n"
		 _gm_version, _gm_build_id));
      _GM_INFO (("Copyright (C) 1999, 2000, Myricom, Inc.\n"));
      printed = 1;

      /* Report if _GM_PRINT is activated. */
      
      _GM_PRINT (1, ("GM debugging printing is activated.\n"));
    }
#endif

#if GM_KERNEL
/* Warnings to be printed only in kernel mode */
#if !GM_ENABLE_SECURITY
  {
    static int security_warning_printed;

    if (!security_warning_printed)
      {
	GM_WARN (("Some security features are disabled in this GM build\n"
		  "to allow non-privileged users to map the network.\n"));
	security_warning_printed = 1;
      }
  }
#endif

#if GM_ENABLE_GALVANTECH_WORKAROUND
  {
    static int galvantech_warning_printed;

    if (!galvantech_warning_printed)
      {
	GM_WARN (("Galvantech SRAM bug checksum testing enabled"
		  " - expect reduced performance.\n"));
	galvantech_warning_printed = 1;
      }
  }
#endif

#if GM_NUM_PORTS != GM_NUM_PORTS_STANDARD
  {
    static int num_ports_warning_printed;

    if (!num_ports_warning_printed)
      {
	GM_WARN (("Driver supports %d, not the standard %d, GM ports\n",
		  GM_NUM_PORTS, GM_NUM_PORTS_STANDARD));
	num_ports_warning_printed = 1;
      }
  }
#endif
#endif /* GM_KERNEL */

  _gm_global_mutex = gm_create_mutex ();
  if (!_gm_global_mutex)
    {
      GM_NOTE (("Could not create gm_global_mutex.\n"));
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  
  status = _gm_call_trace_init ();
  if (status != GM_SUCCESS)
    {
      goto abort_with_global_mutex;
    }
  
  /* Under Windows, intialize winsock. */

  if (
#if defined WIN32 && !GM_KERNEL && GM_BUILDING_GM_LIB
      WSAStartup (MAKEWORD (2, 0), &gm_wsadata) != 0
#else
      0
#endif
      )
    {
      GM_NOTE (("Could not initialize winsock.\n"));
      status = GM_UNTRANSLATED_SYSTEM_ERROR;
      goto abort_with_call_trace_init;
    }

#if GM_KERNEL
  status = __gm_init_local_mutex ();
#else
  status = GM_SUCCESS;
#endif
  if (status != GM_SUCCESS)
    {
      goto abort_with_WSAstartup;
    }
  /* gm_assert (GM_PAGE_LEN != 0); */

  /* initialize GM_PAGE_LEN */

#if !GM_KERNEL
  status = _gm_get_page_len (&GM_PAGE_LEN);
#else
  status = GM_SUCCESS;
#endif
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("Could not determine page length.\n"));
      goto abort_with_local_mutex;
    }

#if GM_OS_MACOSX
  GM_PAGE_LEN_MSC_is_lame ();
#endif  

  /* initialize random number generator */

  gm_srand (42);

  /* report successful initialization */

done:
  gm_mutex_enter (_gm_global_mutex);
  initialized = ++_gm_initialized;
  gm_mutex_exit (_gm_global_mutex);
  GM_PRINT (GM_DEBUG_GM_INIT,
	    ("_gm_initialized = %lu at exit from gm_init()\n", initialized));

#ifdef DELETE_VXWORKS
#if GM_OS_VXWORKS
  /* This is a hook to allow clean gm_exit()s in VxWorks */

  {
    int error;

    error = setjmp (gm_vxworks_fake_exit_env);
    if (error)
      {
	gm_close_all_ports ();
	return error;
      }
  }
#endif
#endif

  gm_always_assert (sizeof (gm_u8_t) == 1);
  gm_always_assert (sizeof (gm_u16_t) == 2);
  gm_always_assert (sizeof (gm_u32_t) == 4);
  gm_always_assert (sizeof (gm_u64_t) == 8);
  /* If this assertion fails, then the definition of gm_size_t is wrong. */
  gm_always_assert (sizeof (gm_size_t) == sizeof (void *));
  
  return GM_SUCCESS;

 abort_with_local_mutex:
#if GM_KERNEL
  __gm_finalize_local_mutex ();
#endif
 abort_with_WSAstartup:
 abort_with_call_trace_init:
  _gm_call_trace_fini ();
 abort_with_global_mutex:
  gm_destroy_mutex (_gm_global_mutex);
 abort_with_nothing:
  return GM_FAILURE;
}

void
gm_finalize ()
{
  int initialized;

  GM_CALLED ();
  GM_PRINT(GM_DEBUG_GM_INIT, ("gm_finalize() entered\n"));

  gm_mutex_enter (_gm_global_mutex);
  initialized = --_gm_initialized;
  gm_mutex_exit (_gm_global_mutex);
  GM_PRINT(GM_DEBUG_GM_INIT,
           ("_gm_initialized reduced to %d at entry to gm_finalize()\n",
            initialized));
#if GM_KERNEL
  gm_always_assert (initialized >= 0);
#endif

  /* escape if not yet time to finalize */
  /* Note: in user case, redundant calls are possible, resulting in negative
   *       value; e.g., if user calls gm_close() and then gm_finalize()
   */
  if (initialized != 0)		/* may be negative if not GM_KERNEL */
    GM_RETURN_NOTHING ();

  /* finalize */

#if GM_KERNEL
  __gm_finalize_local_mutex ();
#endif

  /* HACK to return cleanly after the call trace has been finalized */
  GM_RECORD_RETURN ("%s\n", "");
  _gm_call_trace_fini ();
  
  GM_PRINT(GM_DEBUG_GM_INIT, ("gm_finalize() destroying _gm_global_mutex\n"));
  gm_destroy_mutex (_gm_global_mutex);

  return;
}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
