/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: glenn@myri.com */

/* This file abstracts the LANai hardware packet and DMA engines so that the
   single API presented here can be used to control any LANai. */

#ifndef _GM_bootstrap_h_
#define _GM_bootstrap_h_

#include <stdio.h>

#include "gm_debug_chained_dma.h"
#include "gm_compiler.h"
#include "gm_debug.h"
#include "gm_debug_lanai_dma.h"
#include "gm_dma_chain.h"
#include "gm_enable_crc32.h"
#include "gm_enable_debug_counters.h"
#include "gm_lanai_def.h"
#include "gm_state_inlines.h"
#include "gm_types.h"
#include "gmcp.h"

/* Sparc streaming DMA settings */

#define GM_SPARC_STREAMING (GM_CPU_sparc || GM_CPU_sparc64)
#define GM_SPARC_CACHE_LINE 64
#define GM_SPARC_IO_PAGE_LEN 8192
#define GM_SPARC_STREAMING_FLUSH_LEN 4

/****************************************************************
 * Forward declarations
 ****************************************************************/

static void set_ISR (gm_u32_t);
static gm_u32_t get_ISR (void);

/****************************************************************
 * LANai hardware initialization.
 ****************************************************************/

void gm_init_lanai_hardware (void);

/****************************************************************
 * Abstraction of state bits, including ISR encodings.
 ****************************************************************/

/* Introduce a barrier between a write to a special register and
   the subsequent read of ISR.

   Here we pass SPECIAL as a pointer to prevent the compiler from from
   re-reading the volatile special register after it is written. */

static inline void
isr_barrier (gm_u32_t * special)
{
#if L4 | L5 | L6
  asm ("nop ! ISR_BARRIER()": : "m" (*special):"isr");
#elif L7 | L8
  GM_PARAMETER_MAY_BE_UNUSED (special);
  /* no need to introduce delay */
#elif L9
  asm ("nop ! ISR_BARRIER()": "=m" (ISR):"m" (*special));
#else
#error unrecognized LANai type
#endif
}

#define ISR_BARRIER(special) isr_barrier ((gm_u32_t *) &(special))

/**********************************************************************
 * DMA interface code for LANais
 **********************************************************************/

/****************************************
 * DMA interface code for LANais
 ****************************************/

#include "gm_lanai_def.h"
#include "gm_types.h"
#include "gm_debug.h"
#include "gm_debug_lanai_dma.h"
#include "gm_stbar.h"
#include "gm_gcc_version.h"
#include "gm_lanai_globals.h"

#if L4
/* nothing to do */
#elif L5
extern void *LAR;
extern void *EAR;
extern int DMA_DIR;
#elif L6 | L7 | L8 | L9
/* nothing to do */
#else
#error
#endif

/****************
 * ISR
 ****************/

extern int isr_state_modified_since_get_isr;

static inline gm_u32_t
get_ISR (void)
{
  gm_u32_t ret;

  /* No need for any sort of barrier here.  That is taken care of
     in set_RML() et al. */

#if (L4 | L5 | L6)		/*  && GCC_VERSION >= GCC (2,8) */
  register gm_u32_t isr asm ("isr");
  asm volatile ("nop ! ISR barrier":"=r" (isr));
  ret = isr;
#elif L7 | L8 | L9
  ret = ISR;
#endif
  if (GM_DEBUG)
    {
      isr_state_modified_since_get_isr = 0;
    }
  return ret;
}

static inline void
set_ISR (gm_u32_t val)
{
#if defined lanai3		/*  && GCC_VERSION >= GCC (2,8) */
  /* BAD: circumvent volatile register bug in lanai3-gcc-2.8.1 */
  asm volatile ("mov %0,%?isr\n\t"::"r" (val):"isr", "imr", "memory");
#else
  ISR = (int) val;
  GM_STBAR ();
#endif

  if (GM_DEBUG && (val & IMR_INIT_VAL))
    {
      isr_state_modified_since_get_isr = 1;
    }
}

/****************
 * IMR
 ****************/

static inline gm_u32_t
get_IMR (void)
{
#if L4 | L5 | L6
  register gm_u32_t imr asm ("imr");
  asm volatile ("! IMR barrier":"=r" (imr));
  return imr;
#elif L7 | L8 | L9
  gm_u32_t ret;

  ret = IMR_INIT_VAL;
  return ret;
#else
#error unrecognize processor type
#endif
}

static inline void
set_IMR (gm_u32_t val)
{
#if L4 | L5 | L6
#if GCC_VERSION >= GCC(2,95) || 1
  /* This compiler does not support volatile registers */
  asm volatile ("mov %0,%?imr"::"r" (val):"isr", "imr", "memory");
#else
  IMR = val;
#endif
#elif L7 | L8 | L9
  /* No IMR register on this machine. */
  gm_assert (get_IMR () == val);
#else
#error Do not know what to do.
#endif
  GM_STBAR ();
}

/****************
 * SML, SMLT
 ****************/

/* start a send DMA using the previously set SA and SMP pointers.
   "sml" is rounded down to a GM_PACKET_GRANULARITY boundary, and
   the resulting sml_prime pointer points to the first byte that will not be
   sent.  sml_prime must be larger than the previously set SMP. */

static inline void
set_SML (void *sml)
{
  gm_assert (((void *) (sml) > (void *) gm.send_chunk[0].route &&
	      (void *) (sml) <= (void *) (&gm.send_chunk[0] + 1)) ||
	     ((void *) (sml) > (void *) gm.send_chunk[1].route &&
	      (void *) (sml) <= (void *) (&gm.send_chunk[1] + 1)) ||
	     ((void *) (sml) > (void *) &gm_connection[0] &&
	      (void *) (sml) <= (void *) (&gm_connection
					  [GM_NUM_PRIORITIES *
					   (gm.max_node_id + 1)])));
  gm_assert ((void *) (sml) > (void *) SMP);

  GM_STBAR ();
#if L4
  SML = (void *) ((char *) (sml) - 4);
#elif L5 | L6 | L7 | L8 | L9
  SML = (void *) ((char *) (sml) - 8);
#else
#error
#endif
  ISR_BARRIER (SML);
}

/* Like SML, marks the last flit as a tail flit. */

static inline void
set_SMLT (void *sml)
{
  gm_assert (((void *) (sml) > (void *) gm.send_chunk[0].route &&
	      (void *) (sml) <= (void *) (&gm.send_chunk[0] + 1)) ||
	     ((void *) (sml) > (void *) gm.send_chunk[1].route &&
	      (void *) (sml) <= (void *) (&gm.send_chunk[1] + 1)) ||
	     ((void *) (sml) > (void *) &gm_connection[0] &&
	      (void *) (sml) <= (void *) (&gm_connection
					  [GM_NUM_PRIORITIES *
					   (gm.max_node_id + 1)]))
	     || ((void *) (sml) > (void *) &gm.ethernet.send.chunk[0].packet
		 && (void *) (sml) <=
		 (void *) (&gm.ethernet.send.chunk[0].packet + 1))
	     || ((void *) (sml) > (void *) &gm.ethernet.send.chunk[1].packet
		 && (void *) (sml) <=
		 (void *) (&gm.ethernet.send.chunk[1].packet + 1)));
  gm_assert ((void *) (sml) > (void *) SMP);

  GM_STBAR ();
#if L4
  SMLT = (void *) (((char *) sml + GM_PACKET_GRANULARITY - 1) - 4);
#elif L5 | L6 | L7 | L8 | L9
  /* Do NOT subract 8 for the L5/6 SMLT */
  SMLT = (void *) ((char *) sml + GM_PACKET_GRANULARITY - 1);
#else
#error
#endif
  ISR_BARRIER (SMLT);
}

/* Start a receive using the previously set RMP value.  rml is rounded
   down to a GM_PACKET_GRANULARITY boundary and the resulting rml_prime
   pointer points to the first byte past the end of the receive buffer. */

static inline void
set_RML (void *rml)
{
  gm_assert ((void *) (rml) == (void *) (&gm.recv_chunk[0].end) ||
	     (void *) (rml) == (void *) (&gm.recv_chunk[1].end));
  gm_assert ((void *) RMP == (void *) gm.recv_chunk[0].packet.as_bytes ||
	     (void *) RMP == (void *) gm.recv_chunk[1].packet.as_bytes);
  gm_assert ((void *) (rml) > (void *) RMP);

  GM_STBAR ();
#if L4
  RML = (void *) ((char *) (rml) - 4);
#elif L5 | L6 | L7 | L8 | L9
  RML = (void *) ((char *) (rml) - 8);
#else
#error
#endif
  ISR_BARRIER (RML);
}

static inline void
set_SMH (void *ptr)
{
#if GM_ENABLE_CRC32 && (L7 | L8 | L9)
  if (ptr)
    {
      SMH = (ptr);
    }
#else
  GM_PARAMETER_MAY_BE_UNUSED (ptr);
#endif
}

/****************
 * Host DMA
 ****************/

/* Write a value to the PULSE register */

#if L6 | L7 | L8 | L9
static inline
void
gm_write_PULSE (gm_u32_t value)
{
  /* Make sure the DMA descriptor has been written */
  GM_STBAR ();
  PULSE = value;
  if (L9)
    {
      /* Write the PULSE an extra time on the L9, since some L9 cards
         have a bug where up to 1 out of every million pulses may be
         lost.  Writing the pulse a second time should reduce the rate
         to 1 in 10^12, and only costs one instruction per DMA. */
      PULSE = value;
    }
  /* This should not be necessary, since writing the PULSE does not
     effect the ISR directly, but I leave this here because that's
     what we did before. --Glenn */
  ISR_BARRIER (PULSE);
}
#endif

/* DMA direction macros for use with set_DMA_CTR. */
#if L4
#define GM_SDMA_DIR (-1)
#define GM_RDMA_DIR 0
#elif L5 | L6 | L7 | L8 | L9
#define GM_SDMA_DIR DMA_E2L
#define GM_RDMA_DIR DMA_L2E
#else
#error
#endif

/* DMA_done(): Return 1 if and only iff no DMA is in progress. */

extern int gm_chained_dma_initialized;
extern gm_dma_descriptor_t *gm_chained_dma_free_descriptor;

static inline int
DMA_done (void)
{
#if L4 | L5
#error
#elif L6 | L7 | L8 | L9
  GM_STBAR ();
  if (GM_SPARC_STREAMING)
    {
      return ((gm_u32_t) gm.dma_descriptor.next_with_flags
	      & (gm_u32_t) gm.flush_dma_descriptor.next_with_flags
	      & DMA_TERMINAL);
    }
  else
    {
      return ((gm_u32_t) gm.dma_descriptor.next_with_flags
	      & DMA_TERMINAL);
    }
#else
#error
#endif
}

/* user_DMA_done(): Return nonzero if and only iff no user DMA is in
   progress. */

static inline int
user_DMA_done (void)
{
#if L4 | L5
#error
#elif L6 | L7 | L8 | L9
  gm_dma_descriptor_t *a, *b;
  gm_dma_descriptor_lp_t a__next_with_flags, b__next_with_flags;

  GM_STBAR ();
  a = gm.sdma.tail_dma_descriptor;
  b = gm.rdma.tail_dma_descriptor;
  ;
  a__next_with_flags = a->next_with_flags;
  b__next_with_flags = b->next_with_flags;
  ;
  return (GM_DMA_DESCRIPTOR_TERMINAL
	  & (gm_size_t) a__next_with_flags
	  & (gm_size_t) b__next_with_flags);
#else
#error
#endif
}

/* Spin until no DMA is in progress. */

static inline void
await_free_DMA_engine (void)
{
  /* wait for all DMAs to complete */
  
  while (!DMA_done ())
    ;
}

static inline void
gm_clear_DMA_INT_BIT (void)
{
  set_ISR (DMA_INT_BIT);
}

#define set_EAR(ea, dma_dir) do {					\
  GM_PRINT (GM_TRACE_LANAI_DMA, ("set_EAR(0x%qx,%d) called\n",		\
				      (gm_u64_t) (ea), (dma_dir)));	\
  gm_assert ((ea) >= GM_PAGE_LEN);					\
  _set_EAR ((ea), (dma_dir));						\
} while (0)

static inline void
_set_EAR (gm_dp_t ea, int dma_dir)
{
#if L4 | L6
  GM_PARAMETER_MAY_BE_UNUSED (dma_dir);
  EAR = (gm_lp_t) (((gm_u32_t) ea) & ((gm_u32_t)(~3)));  /* round down to fix SBus dirsend problem */
#elif L5
  if (dma_dir == GM_SDMA_DIR)
    {
      EAR = (gm_lp_t) (gm_u32_t) ea;
      E2L_EAR = (gm_lp_t) (gm_u32_t) ea;
    }
  else
    {
      EAR = (gm_lp_t) (gm_u32_t) ea;
      L2E_EAR = (gm_lp_t) (gm_u32_t) ea;
    }
#elif L7 | L8 | L9
  GM_PARAMETER_MAY_BE_UNUSED (dma_dir);
  if (sizeof (ea) == 8)
    {
      gm_assert (&gm.dma_descriptor.ebus_addr_high + 1
		 == &gm.dma_descriptor.ebus_addr_low);
      asm volatile ("nop	! begin EAR write");
      *(gm_dp_t *) & gm.dma_descriptor.ebus_addr_high = ea;
      asm volatile ("nop	! end EAR write");
      GM_PRINT (GM_TRACE_LANAI_DMA,
		("EAR was set to 0x%qx\n",
		 *(gm_u64_t *) & gm.dma_descriptor.ebus_addr_high));
    }
  else
    {
      gm_assert (gm.dma_descriptor.ebus_addr_high == 0);
      gm.dma_descriptor.ebus_addr_low = ea;
    }
#else
#error Unrecognized LANai version.
#endif
}

static inline gm_dp_t
get_EAR (void)
{
#if L4 | L5 | L6
  return (gm_u32_t) EAR;
#elif L7 | L8 | L9
  return *(gm_dp_t *) & gm.dma_descriptor.ebus_addr_high;
#endif
}

#undef EAR
#define EAR FIXME

static inline void
set_LAR (gm_lp_t la, int dma_dir)
{
#if L4 | L6
  GM_PARAMETER_MAY_BE_UNUSED (dma_dir);
  LAR = la;
#elif L5
  if (dma_dir == GM_SDMA_DIR)
    {
      LAR = la;
      E2L_LAR = la;
    }
  else
    {
      LAR = la;
      L2E_LAR = la;
    }
#elif L7 | L8 | L9
  GM_PARAMETER_MAY_BE_UNUSED (dma_dir);
  gm.dma_descriptor.lanai_addr = la;
#else
#error Unrecognized LANai version.
#endif
}

static inline void
set_LED (unsigned int x)
{
#if defined(CSPI_MAP26xx)
  /* Don't touch the CSPI LED, since it is used to control the host
     CPU reset.*/
#warning CSPI_MAP26xx defined, Lanai LED disabled
#else
  LED = x;
#endif
}

#define GM_SPARC_IO_PAGE_OFFSET(ptr)					\
/* */ GM_MISALIGNMENT (ptr, GM_SPARC_IO_PAGE_LEN)

/* set the DMA_CTR and initiate a DMA.

   COUNT specifies the number of bytes to DMA.  COUNT is rounded down
   to GM_DMA_GRANULARITY before being used.  COUNT must not be zero or
   be rounded down to zero.

   DIR specifies the DMA direction, either GM_SDMA_DIR or GM_RDMA_DIR.

   LIMIT is the DMA address of the byte after the last byte DMAd, or 0
   to disable automatic DMA coherence maintenance.
*/

#define set_DMA_CTR(count, dir, limit) do {				\
  __set_DMA_CTR ((count), (dir), (limit), __GM_FILE__, __LINE__);	\
} while (0)

static inline void
__set_DMA_CTR (unsigned dc, unsigned dma_dir, gm_dp_t limit,
	       const char *file, int line)
{
  GM_PARAMETER_MAY_BE_UNUSED (file);
  GM_PARAMETER_MAY_BE_UNUSED (line);
  
  gm_assert ((dc) > 0);
  gm_assert ((dc) <= GM_MAX_DMA_CTR);
  gm_assert ((dc) >= GM_DMA_GRANULARITY);
  GM_DEBUG_CHAINED_DMA_ASSERT
    (dma_dir == GM_DMA_DESCRIPTOR_L2E || dma_dir == GM_DMA_DESCRIPTOR_E2L);
  
#if GM_DMA_GRANULARITY > 1 && GM_DMA_GRANULARITY_ROUNDUP == 0
  gm_assert (GM_DMA_ALIGNED (dc));
  gm_assert (dc > 0 && dc < GM_MAX_DMA_CTR);
  gm_assert (GM_DMA_ALIGNED (gm.dma_descriptor.ebus_addr_low));
  gm_assert (GM_DMA_ALIGNED (gm.dma_descriptor.lanai_addr));
#endif

#if L4
#error not supported
#elif L6 | L7 | L8 | L9

  /* record DMA length first !!!! the DMA chip might not yet be stopped */

  gm.dma_descriptor.len = dc;

  /* Flush all descriptor updates before chaining. */
  
  GM_STBAR ();
 
  /* record DMA direction while maintaining other settings */

  if (GM_SPARC_STREAMING
      && ((dma_dir == GM_RDMA_DIR
	   && !GM_ALIGNED (limit, GM_SPARC_CACHE_LINE))
	  || (dma_dir == GM_SDMA_DIR
	      && !GM_ALIGNED (limit, GM_SPARC_IO_PAGE_LEN))))
    {
      /* Either the end of the RDMA is not cache line aligned, or the
	 last byte of the SDMA is not in the last cache line of a
	 page.  In the RDMA case, we need to do an SDMA readback in
	 the same page to flush the RDMA.  In the SDMA case, we want
	 to read from the last cache line in the page to leave the
	 prefetch cache in the a known state of having read from the
	 last cache line in the page.  In both cases, we use the DMA
	 chaining hardware to schedule a small SDMA from the very end
	 of the page after the desired DMA. */

      if (sizeof (limit) == 8)
	{
	  gm.flush_dma_descriptor.ebus_addr_high = (limit >> 31) >> 1;
	}
      else
	{
	  gm_assert (gm.flush_dma_descriptor.ebus_addr_high == 0);
	}
      gm_assert (gm.flush_dma_descriptor.len
		 == 2 * GM_SPARC_STREAMING_FLUSH_LEN);
      gm_assert (gm.flush_dma_descriptor.lanai_addr
		 == (gm_lp_t)gm.dummy_flush_target);
      /* read at end of page */
      gm.flush_dma_descriptor.ebus_addr_low
	= (GM_ALIGN (u32, limit, GM_SPARC_IO_PAGE_LEN)
	   + (GM_SPARC_IO_PAGE_LEN 
	      - GM_SPARC_STREAMING_FLUSH_LEN 
	      - GM_SPARC_CACHE_LINE));
      gm.flush_dma_descriptor.next_with_flags
	= (gm_lp_t) (&gm.dma_descriptor) + GM_SDMA_DIR;
      gm.dma_descriptor.next_with_flags
	= (gm_lp_t) (&gm.flush_dma_descriptor) + dma_dir;
    }
  else
    {
      gm.dma_descriptor.next_with_flags
	= (gm_lp_t) ((char *) (&gm.dma_descriptor) + dma_dir);
    }

  gm_assert (GM_DMA_ALIGNED (&gm.dma_descriptor));

  /* Flush descriptor and start DMA */
  gm_write_PULSE (2);
#else
#error
#endif
}


/****************************************************************
 * Start a simple RDMA
 ****************************************************************/

/* In these functions, lar, ear, and len are each rounded down to
   GM_DMA_GRANULARITY before use.  (The rounding usually is done
   by the hardware.) */

#define synchronous_RDMA(lar, ear, len) do {				\
  GM_PRINT (GM_TRACE_LANAI_DMA,						\
		 ("_start_RDMA(0x%p,0x%qx,0x%x) called\n",		\
		  lar, (gm_u64_t) ear, len));				\
  gm_assert (len >= GM_DMA_GRANULARITY);				\
  _synchronous_RDMA (lar, ear, len);					\
} while (0)

#define GM_ENABLE_RDMA_READBACK 0

static inline void
_synchronous_RDMA (gm_lp_t lar, gm_dp_t ear, gm_u32_t len)
{
  GM_STBAR ();
  set_LAR (lar, GM_RDMA_DIR);
  set_EAR (ear, GM_RDMA_DIR);
  set_DMA_CTR (len, GM_RDMA_DIR, ear + len);
  await_free_DMA_engine ();

  if (GM_ENABLE_RDMA_READBACK && (L7 || L9))
    {
      GM_NOTE_OCCASIONALLY
	(("GM_ENABLE_RDMA_READBACK is slow and for testing only\n"));
      
      set_LAR (&gm.trash, GM_SDMA_DIR);
      set_EAR (ear+len-1, GM_SDMA_DIR);
      set_DMA_CTR (1, GM_SDMA_DIR, 0);
      await_free_DMA_engine ();
    }
}

/* start a simple RDMA into the recv queue */

#define synchronous_GRANULAR_RDMA(lar, ear, len) do {			\
  gm_assert (GM_RDMA_ALIGNED (ear));					\
  gm_assert (GM_RDMA_ALIGNED (len));					\
  synchronous_RDMA (lar, ear, len);					\
} while (0)

/* start a simple SDMA.  The start of the DMA must be aligned, but the
   length may be misaligned, in which case a bit extra will be DMAd. */

#define synchronous_SDMA(ear, lar, len) do {				\
  gm_assert ((len) >= GM_DMA_GRANULARITY);				\
  _synchronous_SDMA ((ear), (lar), (len));				\
} while (0)

static inline void
_synchronous_SDMA (gm_dp_t ear, gm_lp_t lar, gm_u32_t len)
{
#if GM_SPARC_STREAMING
  /* Check for the defeat of our streaming DMA consistency strategy. */
  
  if (GM_SPARC_IO_PAGE_OFFSET (ear)
      >= GM_SPARC_IO_PAGE_LEN - 3 * GM_SPARC_CACHE_LINE)
    {
      /* The SDMA starts in the last cache line of a page.  This is
	 known to be the same cache line as the last cache line we
	 SDMA'd from the page, if any.  Therefore, the Sparc prefetch
	 cache might hold cached data for that cache line and we must
	 flush it to maintain DMA consistency.  We flush the cache by
	 SDMAing a bit of data from the start of the page. */
      
      set_LAR (gm.dummy_flush_target, GM_SDMA_DIR);
      /* perform a read DMA that touches the first 2 cache lines in the
	 page. */
      set_EAR (ear - GM_SPARC_IO_PAGE_OFFSET (ear)
	       + (GM_SPARC_CACHE_LINE - GM_SPARC_STREAMING_FLUSH_LEN),
	       GM_SDMA_DIR);
      set_DMA_CTR (GM_SPARC_STREAMING_FLUSH_LEN * 2, GM_SDMA_DIR, 0);
      await_free_DMA_engine ();
    }
#endif
  set_LAR (lar, GM_SDMA_DIR);
  set_EAR (ear, GM_SDMA_DIR);
  GM_STBAR ();
  set_DMA_CTR (len, GM_SDMA_DIR, ear + len);
  await_free_DMA_engine ();
}

/* set the interrupt timer */

static inline void
set_IT (signed it)
{
  GM_STBAR ();
#if L4 | L5
  IT = (it);
  ISR_BARRIER (IT);
#elif L6 | L7 | L8 | L9
  IT0 = (it);
  ISR_BARRIER (IT0);
#else
#error
#endif
}

/****************************************************************
 * Parity checking
 ****************************************************************/

static inline void
gm_check_parity (void)
{
  int error = 0;

#if L7
  error = ((get_ISR () & PARITY_INT_BIT) != 0);
#elif L8 | L9
  error = ((get_ISR () & PAR_INT_BIT) != 0);
#endif

  if (error)
    {
      gm_printf ("LANai detected a PARITY error in board SRAM"
		" at rtc time = 0x%x%08x\n", *(gm_u32_t *) 4, *(gm_u32_t *) 0);
      fflush (stdout);
      gm_always_assert (0);
      while (1) {
	/* freeze the mcp */
      }
    }
}

/* On some interfaces, PULSE signals can be lost rarely.  (The worst
   observed NIC dropped less that 2 pulses per million.) On these NICs
   generate an extra PULSE signal periodically to ensure that the PCI
   DMA chip continues to poll the descripters and recovers from this
   rare case. */

static inline
void
gm_generate_defensive_pulse (void)
{
#if L9
  /* Trigger the DMA engine. */
  gm_write_PULSE (6);
#endif
}

/****************************************************************
 * 64-bit timer
 ****************************************************************/

extern union __rtc64
{
  gm_s64_t whole;
  struct
  {
    gm_s32_t high;
    gm_u32_t low;
  }
  parts;
}
__rtc64;

/* This function must be called at least once per clock rollover for
   correct operation.  Therefore, it is called in the low-frequency
   timeout code in gm_timer.h. */

static inline gm_s64_t
rtc64 (void)
{
  gm_u32_t rtc;

  rtc = RTC;
  if (rtc < __rtc64.parts.low)
    __rtc64.parts.high++;
  __rtc64.parts.low = rtc;

  return __rtc64.whole;
}

#endif /* _gm_bootstrap_h_ defined */

/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  c-backslash-column:72
  tab-width:8
  End:
*/
