/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_myrinet_packet_descriptor.h"
#include "gm_debug_mpds.h"
#include "gm_state.h"

gm_status_t gm_mpd_create (gm_myrinet_packet_descriptor_free_list_t *fl,
			   unsigned long mtu)
{
  gm_myrinet_packet_descriptor_t *d;
  unsigned long alloc_len;
  
  mtu = GM_ROUNDUP (u32, mtu, 8);
  alloc_len = (sizeof (gm_myrinet_packet_descriptor_t)
	       + GM_MAX_NETWORK_DIAMETER + mtu + 8);
  d = (gm_myrinet_packet_descriptor_t *) gm_calloc (1, alloc_len);
  if (!d)
    {
      return GM_OUT_OF_MEMORY;
    }
  
  d->free_list = fl;
  d->route.ptr = (char *) (d + 1) + GM_MAX_NETWORK_DIAMETER;
  d->route.limit = d->route.ptr;
  d->header.ptr = d->route.limit;
  d->header.limit = (char *) (d + 1) + alloc_len;
  d->payload.ptr = 0;
  
  gm_mpd_free (d);
  
  return GM_SUCCESS;
}

static gm_myrinet_packet_descriptor_free_list_t *debug_free_lists;

void
gm_mpd_free_list_finalize (gm_myrinet_packet_descriptor_free_list_t *fl)
{
  gm_myrinet_packet_descriptor_t *d, *next;

  /* free all the descriptors */

  d = fl->first_free;
  while (d)
    {
      next = d->next;
      gm_free (d);
      d = next;
    }

  /* If debugging, remove the free list from the list of free lists */

  if (GM_DEBUG_MPD_FREE_LISTS)
    {
      GM_LIST_REMOVE (gm_myrinet_packet_descriptor_free_list_t,
		      debug_next, &debug_free_lists, fl);
    }
}

gm_status_t
gm_mpd_free_list_init (gm_myrinet_packet_descriptor_free_list_t *fl,
		       unsigned cnt, unsigned mtu, gm_u32_t nonempty_state_bit)
{
  gm_status_t status;
  
  gm_bzero (fl, sizeof (*fl));
  
  fl->nonempty_state_bit = nonempty_state_bit;
  
  while (cnt--)
    {
      status = gm_mpd_create (fl, mtu);
      if (status != GM_SUCCESS)
	{
	  goto abort_with_descriptors;
	}
    }

  if (GM_DEBUG_MPD_FREE_LISTS)
    {
      fl->debug_next = debug_free_lists;
      debug_free_lists = fl;
    }
  
  return GM_SUCCESS;
  
 abort_with_descriptors:
  gm_mpd_free_list_finalize (fl);
  return status;
}

void
gm_mpd_debug_report (void)
{
  gm_myrinet_packet_descriptor_free_list_t *fl;

  for (fl=debug_free_lists; fl; fl=fl->debug_next)
    {
      _GM_PRINT (GM_DEBUG_MPD_FREE_LISTS,
		 ("mpd %p has %d free entries and notices the FOLLOWING:\n",
		  fl, fl->debug_free_cnt));
      _print_state (fl->nonempty_state_bit);
    }
}

void
_gm_mpd_print (gm_myrinet_packet_descriptor_t *d)
{
  GM_PRINT (GM_DEBUG_MPDS, ("mpd %p = {\n", d));
  _GM_PRINT (GM_DEBUG_MPDS, ("  free_list = %p\n", d->free_list));
  _GM_PRINT (GM_DEBUG_MPDS, ("  next = %p\n", d->next));
  _GM_PRINT (GM_DEBUG_MPDS, ("  route = {\n"));
  _GM_PRINT (GM_DEBUG_MPDS, ("    ptr = %p\n", d->route.ptr));
  _GM_PRINT (GM_DEBUG_MPDS, ("    limit = %p\n", d->route.limit));
  _GM_PRINT (GM_DEBUG_MPDS, ("  }\n"));
  _GM_PRINT (GM_DEBUG_MPDS, ("  header = {\n"));
  _GM_PRINT (GM_DEBUG_MPDS, ("    ptr = %p\n", d->header.ptr));
  _GM_PRINT (GM_DEBUG_MPDS, ("    limit = %p\n", d->header.limit));
  _GM_PRINT (GM_DEBUG_MPDS, ("  },\n"));
  _GM_PRINT (GM_DEBUG_MPDS, ("  payload = {\n"));
  _GM_PRINT (GM_DEBUG_MPDS, ("    ptr = %p\n", d->payload.ptr));
  _GM_PRINT (GM_DEBUG_MPDS, ("    limit = %p\n", d->payload.limit));
  _GM_PRINT (GM_DEBUG_MPDS, ("  },\n"));
  _GM_PRINT (GM_DEBUG_MPDS, ("}\n"));

  if (GM_DEBUG_MPDS)
    {
      _GM_PRINT (GM_DEBUG_MPDS, ("route = \n"));
      gm_hex_dump (d->route.ptr, d->route.limit - d->route.ptr);
      _GM_PRINT (GM_DEBUG_MPDS, ("header = \n"));
      gm_hex_dump (d->header.ptr, d->header.limit - d->header.ptr);
      _GM_PRINT (GM_DEBUG_MPDS, ("payload = \n"));
      gm_hex_dump (d->payload.ptr, d->payload.limit - d->payload.ptr);
    }
  
  if (GM_DEBUG_MPDS)
    {
      fflush (stdout);
    }
}

gm_status_t
gm_mpd_init (void)
{
  gm_status_t status;
  static int initialized;
  unsigned int ui;

  GM_CALLED ();
  
  if (initialized)
    {
      GM_RETURN_STATUS (GM_SUCCESS);
    }

  /****************
   * Send descriptors
   ****************/

  /* initialize the send descriptors */

  status = gm_mpd_free_list_init (&gm.send.free_list, 4, GM_MYRINET_MTU,
				  FREE_SEND_CHUNK);
  if (status != GM_SUCCESS)
    {
      goto abort_with_nothing;
    }
  
  /* initialize the send list */
  
  gm.send.pending.nonempty_state_bit = SEND_PENDING;
  gm.send.pending.first = 0;
  gm.send.pending.last
    = ((gm_myrinet_packet_descriptor_t *)
       ((char *) &gm.send.pending.first
	- GM_OFFSETOF (gm_myrinet_packet_descriptor_t, next)));

  /****************
   * Recv descriptors
   ****************/

  /* initialize the recv descriptors */

#warning debug hack (40)
  status = gm_mpd_free_list_init (&gm.rdma.free_list, 40, GM_MYRINET_MTU,
				  FREE_RECV_CHUNK);
  if (status != GM_SUCCESS)
    {
      goto abort_with_send_free_list;
    }
  
  /* initialize the recv list */
  
  gm.rdma.pending.nonempty_state_bit = RDMA_PENDING;
  gm.rdma.pending.first = 0;
  gm.rdma.pending.last
    = ((gm_myrinet_packet_descriptor_t *)
       ((char *) &gm.rdma.pending.first
	- GM_OFFSETOF (gm_myrinet_packet_descriptor_t, next)));

  /****************
   * Mapper reply initialization
   ****************/

  gm.mapper_state.free_reply_descriptors.nonempty_state_bit = 0;
  gm.mapper_state.free_reply_descriptors.first_free
    = &gm.mapper_state._reply_stage[0].descriptor;

  /* initialize the mapper reply descriptors */
  for (ui = 0; ui < GM_NUM_ELEM (gm.mapper_state._reply_stage); ui++)
    {
      gm_myrinet_packet_descriptor_t *d;

      GM_PRINT (GM_DEBUG_MPDS, ("Initializing mapper reply stage 0x%x\n", ui));
      
      d = &gm.mapper_state._reply_stage[ui].descriptor;
      d->free_list = &gm.mapper_state.free_reply_descriptors;
      d->next = &gm.mapper_state._reply_stage [ui + 1].descriptor;
      d->route.limit
	= d->header.ptr
	= &gm.mapper_state._reply_stage[ui].payload;
      d->payload.ptr = 0;
    }
  gm.mapper_state._reply_stage[ui-1].descriptor.next = 0;
  
  initialized = 1;

  GM_RETURN_STATUS (GM_SUCCESS);

 abort_with_send_free_list:
  gm_mpd_free_list_finalize (&gm.send.free_list);
 abort_with_nothing:
  GM_RETURN_STATUS (status);
}

