/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1996, 1997, 1998, 1999, 2000 by Myricom, Inc.		 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/*************************************
 * Recv Token Hash Table
 *************************************/

#ifndef _gm_recv_token_hash_h_
#define _gm_recv_token_hash_h_

#include "gmcp.h"
#include "gm_debug.h"
#include "gm_enable_debug_counters.h"

static inline const unsigned
rt_hash (gm_recv_token_key_t key)
{
  gm_u32_t ret = key.whole;

  gm_assert (GM_POWER_OF_TWO (GM_RECV_TOKEN_HASH_BINS));

  ret = ret ^ ret >> 11 ^ ret >> 22;
  ret ^= ret >> 5;
  ret &= GM_RECV_TOKEN_HASH_BINS - 1;
  return ret;
}

static inline gm_recv_token_key_t
recv_token_key (unsigned int sender_node_id,
		unsigned int sender_subport_id,
		unsigned int target_subport_id)
{
  gm_recv_token_key_t ret;

  ret.whole = ((sender_node_id << 16)
	       | (sender_subport_id << 8) | (target_subport_id));
  return ret;
}

static inline void
recv_token_hash_insert (gm_recv_token_t * token, unsigned sender_node_id,
			unsigned sender_subport_id,
			unsigned target_subport_id)
{
  gm_recv_token_t **bin, *_bin;
  gm_recv_token_key_t key;

  key = recv_token_key (sender_node_id, sender_subport_id, target_subport_id);
  bin = &gm.recv_token_bin[rt_hash (key)];
  _bin = *bin;
  token->key = key;
  *bin = token;
  token->next = _bin;

  GM_INCR_DEBUG_CNT (gm.hashed_token_cnt);
}

static inline gm_recv_token_t *
recv_token_hash_get (unsigned sender_node_id, unsigned sender_subport_id,
		     unsigned target_subport_id)
{
  gm_recv_token_t *t;
  gm_recv_token_key_t key;

  key = recv_token_key (sender_node_id, sender_subport_id, target_subport_id);

  t = gm.recv_token_bin[rt_hash (key)];
  while (t && t->key.whole != key.whole)
    t = t->next;
  if (GM_DEBUG && !t)
    {
      gm_printf (GM_STR
		 ("no recv token for node 0x%x subport 0x%x"
		  " target_subport 0x%x\n"),
		 sender_node_id, sender_subport_id, target_subport_id);
    }
  return t;
}

static inline gm_recv_token_t *
recv_token_hash_remove (unsigned sender_node_id, unsigned sender_subport_id,
			unsigned target_subport_id)
{
  gm_recv_token_t **holder, *token, *next_token;
  gm_recv_token_key_t token__key, key;

  key = recv_token_key (sender_node_id, sender_subport_id, target_subport_id);
  /* pre */ holder = &gm.recv_token_bin[rt_hash (key)];
  /* pre */ token = *holder;
  ;
  ;
  /* pre */ token__key.whole = token->key.whole;

  while (token)
    {
      next_token = token->next;
      if (token__key.whole == key.whole)
	{
	  /* fill */ token__key.whole = next_token->key.whole;
	  GM_INCR_DEBUG_CNT (gm.hashed_token_cnt);
	  *holder = next_token;
	  return token;
	}
      token__key.whole = next_token->key.whole;
      holder = &token->next;
      token = next_token;
    }
  if (GM_DEBUG)
    gm_printf (GM_STR
	    ("no recv token for node 0x%x subport 0x%x target_subport 0x%x\n"),
	    sender_node_id, sender_subport_id, target_subport_id);
  return 0;
}

static inline gm_recv_token_t *
recv_token_hash_verify (unsigned sender_node_id, unsigned sender_subport_id,
			unsigned target_subport_id)
{
  gm_recv_token_t *t;
  gm_recv_token_key_t key;

  key = recv_token_key (sender_node_id, sender_subport_id, target_subport_id);

  t = gm.recv_token_bin[rt_hash (key)];
  while (t && t->key.whole != key.whole)
    t = t->next;
  return t;
}

/* Scan the hash table, removing any entry for PORT. */

#define GM_DEBUG_RECV_TOKEN_HASH_REMOVE_PORT_REFERENCES 0

static inline void
recv_token_hash_remove_port_references (unsigned int port)
{
  gm_recv_token_t **holder, *t;
  int i;

  if (GM_DEBUG_RECV_TOKEN_HASH_REMOVE_PORT_REFERENCES)
    {
      gm_printf (GM_STR ("entered " __FUNCTION__ "\n"));
      fflush (stdout);
    }

  /* for each bin */
  for (i = 0; i < GM_RECV_TOKEN_HASH_BINS; i++)
    {
      /* Scan for matches */
      holder = &gm.recv_token_bin[i];
      while (*holder)
	{
	  t = *holder;

	  /* If entry is for PORT */
	  if (GM_SUBPORT_PORT (t->key.parts.target_subport_id) == port)
	    {
	      /* remove and free entry */
	      *holder = t->next;
	      t->next = gm_port[port].free_recv_tokens;
	      gm_port[port].free_recv_tokens = t;
	      GM_DECR_DEBUG_CNT (gm.hashed_token_cnt);
	    }
	  else
	    {
	      /* Try next in bin */
	      holder = &t->next;
	    }
	}
    }

  if (GM_DEBUG_RECV_TOKEN_HASH_REMOVE_PORT_REFERENCES)
    {
      gm_printf (GM_STR ("leaving " __FUNCTION__ "\n"));
      fflush (stdout);
    }
}

#endif /* _gm_recv_token_hash_h_ */
