/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm.h"
#include "gm_debug_node_id.h"
#include "gm_dispatch.h"
#include "gm_heartbeat.h"
#include "gm_interrupt.h"
#include "gm_page_hash.h"
#include "gm_port.h"
#include "gm_recv_queue.h"
#include "gm_send_tokens.h"
#include "gm_types.h"

#define GM_DEBUG_TIMER 0
#define GM_DEBUG_TIMER_PRINT 0

/****************************************************************
 * Timer state machine
 ****************************************************************/

void
timer (void)
{
  GM_HANDLER_CALLED ();
  
  /************
   * Handle high-frequency timer events.
   ************/

  /* update the user-readable real-time clock at SRAM address 0. */

  {
    gm_s64_t s64;
    gm_u32_t high, low;
    
    s64 = rtc64 ();
    low = (gm_u32_t) s64;
    high = (gm_u32_t) (s64 >> 32);

    /* Write 64-bit counter in wrong-endian order to address 0,
       writing the high-part first.  For the reason, see the comments
       in gm_ticks(). */
    
    *(gm_u32_t *)4 = high;
    GM_STBAR ();
    *(gm_u32_t *)0 = low;
    GM_STBAR ();
  }

  if ((gm.timeout_time - RTC) <= 0)
    {
      /************
       * Handle low-frequency timer events.
       ************/

      /* Make sure the PCI DMA engine has not stalled due to a rare
	 lost PULSE signal.  This event is so rare that recovering in
	 the low frequency timer should be sufficient. */

      gm_generate_defensive_pulse ();
      
      /* If debugging, print something about once a second. */

      if (GM_DEBUG_TIMER)
	{
	  static gm_u32_t last_second;

	  if (last_second != (RTC & 0xffe00000))
	    {

	      if (GM_DEBUG_TIMER)
		{
	      	  if (gm.first_port_with_alarm)
		    {
		      LOG_DISPATCH
			(0, "A nominal second has passed - alarm pending");
		    }
		  else
		    {
		      LOG_DISPATCH
			(0, "A nominal second has passed.");
		    }
		}

	      if (GM_DEBUG_TIMER_PRINT)
		{
	          fflush (stdout);
	          gm_printf ("A nominal second has passed.\n");
		  gm_printf ("Recent DMA addresses: 0x%x%08x 0x%x%08x\n",
			     gm.dma_descriptor.ebus_addr_high,
			     gm.dma_descriptor.ebus_addr_low,
			     gm.flush_dma_descriptor.ebus_addr_high,
			     gm.flush_dma_descriptor.ebus_addr_low);
	          print_state ();
	          fflush (stdout);
		}
	      last_second = (RTC & 0xfff00000);
	      gm_always_assert (gm.event_index_table_crc
				== gm_crc (_gmp.event_index,
					   sizeof (_gmp.event_index)));
	    }
	} /* GM_DEBUG_TIMER */

      if (GM_DEBUG_NODE_ID)
	{
	  static int this_node_id_saved = 0;

	  /* check the node_id to see if it changed */
	  if (!this_node_id_saved && gm.this_node_id) 
	    {
	      gm_s64_t now = rtc64 ();
	      this_node_id_saved = gm.this_node_id;
	      LOG_DISPATCH (0, "My gmID was set");
	      printf ("My gmID was set to %d   RTC=0x%x%08x\n",
		      gm.this_node_id, 
		      *(gm_u32_t *)4, *(gm_u32_t *)0);
	      fflush (stdout);
	    }

	  if (this_node_id_saved && 
	      (this_node_id_saved != gm.this_node_id))
	    {
	      gm_s64_t now = RTC;
	      LOG_DISPATCH (0,"My gmID changed");
	      printf ("My gmID changed to %d previous=%d   RTC=0x%x%08x\n",
		      gm.this_node_id, this_node_id_saved, 
		      *(gm_u32_t *)4, *(gm_u32_t *)0);
	      fflush (stdout);
	      this_node_id_saved = gm.this_node_id;
	    }
	}
      
      /* check for parity errors */

      gm_check_parity ();

      gm_assert (gm.send_chunk[0].cafebabe == 0xcafebabe);
      gm_assert (gm.send_chunk[1].cafebabe == 0xcafebabe);
      gm_assert (gm.recv_chunk[0].cafebabe == 0xcafebabe);
      gm_assert (gm.recv_chunk[1].cafebabe == 0xcafebabe);
      gm_assert (gm.ethernet.send.chunk[0].cafebabe == 0xcafebabe);
      gm_assert (gm.ethernet.send.chunk[1].cafebabe == 0xcafebabe);
  

      /* print resend_cnt, but only if it changed */

      if (0 && GM_ENABLE_ERROR_COUNTERS)
	{
	  static gm_u32_t resend_cnt;

	  if (resend_cnt != GM_READ_ERROR_CNT (gm.resend_cnt))
	    {
	      resend_cnt = GM_READ_ERROR_CNT (gm.resend_cnt);
	      gm_printf ("0x%x resends\n", resend_cnt);
	      fflush (stdout);
	    }
	}

      /* re-interrupt if an interrupt was not handled. */

      if ((get_ISR () & HOST_SIG_BIT) == 0 &&
	  gm.interrupt.type != GM_NO_INTERRUPT)
	{
	  static unsigned int count;

	  count++;
	  if (GM_POWER_OF_TWO (count))
	    {
	      GM_INFO (("found lost interrupt %u\n", count));
	    }
	}

      /* Decay the backlog decay timer in case some senders defect from the
         queue for any reason. */

      if (1)
	{
	  gm.backlog_delay -= gm.backlog_delay >> 3;
	}
      else
	{
	  if (gm.backlog_delay > GM_SEND_TIMEOUT / 8)
	    gm.backlog_delay -= GM_SEND_TIMEOUT / 8;
	  else
	    gm.backlog_delay = 0;
	}

      /************
       * Alarms
       ************/

      /* Generate any scheduled alarm */

      if (gm.first_port_with_alarm)
	{
	  gm_port_protected_lanai_side_t **port;

	  port = &gm.first_port_with_alarm;
	  do
	    {
	      if ((RTC - (*port)->alarm_time) > 0)
		{
		  /* Hack: pass send queue slot token for the alarm
		     back to the host before the alarm event.  This is
		     necessary to ensure that alarm events take no
		     more than 1 send queue slot at a time. */

		  if (GM_DEBUG_TIMER)
		    {
		      if ((RTC - (*port)->alarm_time) > 20000)
			{
			  LOG_DISPATCH (0, "processing LATE   alarm");
			}
		      else
			{
			  LOG_DISPATCH (0, "processing timely alarm");
			}
		    }
		   
		  update_host_sent_queues ();

		  gm_report (*port, GM_NEW_ALARM_EVENT);
		  (*port)->alarm_set = 0;
		  *port = (*port)->next_with_alarm;
		  LOG_DISPATCH (0, "done processing alarm");
		}
	      else
		{
		  port = &(*port)->next_with_alarm;
		}
	      gm_morse_async ("4 ");
	    }
	  while (*port);
	}

      /************
       * Host synchronization
       ************/

      /* If the Host has requested exclusive access to the LANai memory data
         structures, grant it. */

      if (gm.pause_rqst)
	{
	  LOG_DISPATCH (0, "LANai pausing");
	  if (GM_DEBUG_TIMER)
	    {
	      if (GM_DEBUG_TIMER_PRINT)
		{
	          gm_printf (GM_STR ("pausing\n"));
	          fflush (stdout);
		}
	    }
	  gm.pause_cnt++;

	  /* Wait for any previous host interrupt to complete before
	     queueing another interrupt.  Also wait for any DMA to
	     complete, because only then is the LANai really paused. */

	  prepare_to_interrupt ("1 ");
	  await_user_DMA_completion ();

	  gm_interrupt (GM_PAUSE_INTERRUPT);

	  /* Flush the cache because it might be updated during the pause. */

#if GM_ENABLE_VM
	  flush_page_hash_cache ();
#endif

	  /* Wait to resume. */

	  prepare_to_interrupt ("2 ");
	  while (gm.pause_rqst)
	    {
	    }

	  /* Ack the resume request. */

	  LOG_DISPATCH (0, "LANai UNpausing");
	  gm_interrupt (GM_UNPAUSE_INTERRUPT);
	  if (GM_DEBUG_TIMER_PRINT)
	    {
	      gm_printf (GM_STR ("unpaused\n"));
	      fflush (stdout);
	    }

	  /* Check if the driver has requested that a port be closed. */

	  if (gm.port_to_close != -1)
	    {
	      gm_always_assert (gm.port_to_close < GM_NUM_PORTS);
	      LOG_DISPATCH (0, "resetting port");
	      if (GM_DEBUG_TIMER_PRINT)
		{
		  gm_printf (GM_STR ("resetting port\n"));
		  fflush (stdout);
		}
	      reset_port (gm.port_to_close);
	      gm.port_to_close = -1;
	      LOG_DISPATCH (0, "reset port");
	      if (GM_DEBUG_TIMER_PRINT)
		{
		  gm_printf (GM_STR ("reset port\n"));
		  fflush (stdout);
		}
	    }
	  if (gm.port_to_open != -1)
	    {
	      gm_always_assert (gm.port_to_close < GM_NUM_PORTS);
	      LOG_DISPATCH (0, "opening port");
	      if (GM_DEBUG_TIMER_PRINT) 
		{
		  gm_printf (GM_STR ("opening port\n"));
		}
	      open_port (&gm_port[gm.port_to_open]);
	      gm.port_to_open = -1;
	      LOG_DISPATCH (0, "opened port");
	      if (GM_DEBUG_TIMER_PRINT)
		{
		  gm_printf (GM_STR ("opened port\n"));
		  fflush (stdout);
		}
	    }
	}

      /* Schedule next timeout and timer event. */

      gm.timeout_time = RTC + GM_LOW_FREQ_TIMER_PERIOD;
    }

  /* flash the LED */
  heartbeat ();

  /* DMA the aggregated lists of sent messages to the host, or
     arrange to do so when the current DMA completes. */

  if (gm.first_port_with_sent_packets)
    {
      update_host_sent_queues ();
    }


  /* Schedule next timer interrupt */
  set_IT (GM_TIMER_PERIOD);

  DISPATCH (132, "");
}

/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  c-backslash-column:72
  tab-width:8
  End:
*/
