/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_user_dma_h_
#define _gm_user_dma_h_

#include "gm_chained_dma.h"
#include "gm_debug_lanai_tlb.h"
#include "gm_enable_directed_send.h"
#include "gm_page_hash.h"
#include "gm_pte.h"

/* Don't inline the functions in this file if debugging.  This saves lots
   of space in the MCP. */

#if GM_DEBUG
#define inline
#endif

/**********************************************************************/
/* Page table stuff */

#if GM_ENABLE_VM

#define GM_DEBUG__DMA_ADDR 0
#define GM_DEBUG_DMA_DESCRIPTORS 0

/* Convert a host virtual address associated with a port to a physical
   address. Returns default plus the page offset if no such page exists. */

static inline gm_dp_t
_dma_addr (gm_up_t host_address, unsigned int port, gm_dp_t default_addr)
{
  gm_cached_pte_t *entry;
  gm_up_t page_port;
  unsigned int offset;
  gm_dp_t ret;
  
  if (GM_DEBUG__DMA_ADDR)
    {
      gm_printf ("_dma_addr(0x%qx,%d,0x%qx) called.\n",
		 (gm_u64_t) host_address, port, (gm_u64_t) default_addr);
    }

  /* preconditions */

  gm_assert (gm.page_hash.cache.root.older);
  gm_assert (gm.page_hash.cache.root.younger);
  gm_assert (default_addr);
  gm_assert (GM_PAGE_ALIGNED (default_addr));
  gm_assert (GM_POWER_OF_TWO (GM_MAX_PAGE_HASH_CACHE_INDEX + 1));
  gm_assert (gm.page_hash.cache.entry
	     [GM_MAX_PAGE_HASH_CACHE_INDEX + 1].pte.page_port == 0);
  gm_assert (port < GM_NUM_PORTS);

  offset = GM_PAGE_OFFSET (host_address);
  page_port = host_address - offset + port;

  /* pre */ entry = (&gm.page_hash.cache.entry
		     [GM_HASH_PAGE_PORT (page_port)
		      & GM_MAX_PAGE_HASH_CACHE_INDEX]);
  if (!page_port)
    {
      if (GM_DEBUG__DMA_ADDR)
	{
	  gm_pte_print (GM_STR ("first pte empty"), &entry->pte);
	  gm_printf (GM_STR (__GM_FUNCTION__ " returning default_addr\n"));
	}
      return default_addr;
    }

  /* Scan for the cached page table entry. */

  if (GM_DEBUG__DMA_ADDR)
    {
      gm_printf (GM_STR ("starting to scan for PTE\n"));
      fflush (stdout);
    }

  while (1)
    {
      gm_up_t entry__page_port;

      if (GM_DEBUG__DMA_ADDR)
	{
	  gm_pte_print (GM_STR ("first pte empty"), &entry->pte);
	  gm_printf (GM_STR ("scanning for PTE\n"));
	  fflush (stdout);
	}

      /* verify that this table is followed by and entry with a user page
         of zero. */
      gm_assert (
		 (gm.page_hash.cache.
		  entry[GM_MAX_PAGE_HASH_CACHE_INDEX + 1].pte.page_port) ==
		 0);

      entry__page_port = entry->pte.page_port;
      if (entry__page_port == page_port)
	{
	  /* Cache hit! */

	  gm_cached_pte_t *younger, *older, *root, *youngest;
	  gm_dp_t entry__dma_page;

	  GM_INCR_DEBUG_CNT (gm.hit_cnt);
	  /* gm_morse_sync ("hit\n"); */

	  /* move the entry to the head of the age queue.  This code
	     is necessarily ugly to pipeline perfectly on the L7. */

	  /* Remove entry from age queue. */

	  younger = entry->younger;
	  older = entry->older;
	  /* pre */ root = &gm.page_hash.cache.root;
	  /* pre */ entry->younger = root;
	  younger->older = older;
	  older->younger = younger;

	  /* Insert entry at head of age queue. */

	  youngest = root->older;
	  root->older = entry;
	  /* pre */ entry__dma_page = gm_pte_get_dma_page (&entry->pte);
	  entry->older = youngest;
	  youngest->younger = entry;

	  /* Return DMA address. */

	  ret = entry__dma_page + offset;
	  break;
	}
      else if (entry__page_port == 0)
	{
	  gm_dp_t dma_page;

	  /* Check for wraparound */

	  if (entry > &gm.page_hash.cache.entry[GM_MAX_PAGE_HASH_CACHE_INDEX])
	    {
	      entry = &gm.page_hash.cache.entry[0];
	      continue;
	    }

	  /* Cache miss.  Fetch entry from host. */

	  GM_INCR_DEBUG_CNT (gm.miss_cnt);
	  /* gm_morse_sync ("miss\n"); */
	  dma_page = get_uncached_dma_page_entry (page_port);
	  if (GM_DEBUG_LANAI_TLB && !dma_page)
	    {
	      GM_PANIC (
			("page_port=0x%qx port=%d host_address=0x%qx\n",
			 (gm_u64_t) page_port, port,
			 (gm_u64_t) host_address));
	    }

	  /* Determine return value. (dma_page is now the
	     corresponding DMA address, or zero if no match.) */

	  if (dma_page)
	    {
	      ret = dma_page + offset;
	      if (GM_DEBUG__DMA_ADDR)
		{
		  gm_printf (GM_U64_TMPL" + 0x%x = "GM_U64_TMPL"\n",
			     GM_U64_ARG(dma_page), offset,
		             GM_U64_ARG(ret));
		}
	    }
	  else
	    {
	      if (GM_DEBUG_LANAI_TLB)
		{
		  gm_pte_print (GM_STR ("first empty pte"), &entry->pte);
		  gm_printf (GM_STR ("no match; returning default\n"));
		}
	      ret = default_addr + offset;
	    }
	  break;
	}
      entry++;
    }

  /* return match */

  if (GM_DEBUG__DMA_ADDR)
    {
      gm_pte_print (GM_STR ("matching pte"), &entry->pte);
      gm_printf (GM_STR ("scanning complete, returning 0x%qx\n"),
		 (gm_u64_t) ret);
      fflush (stdout);
    }

  return ret;
}
#endif /* GM_ENABLE_VM */

static inline gm_dp_t
sdma_addr (gm_up_t host_address, unsigned int port)
{
  GM_PRINT (GM_DEBUG_LANAI_TLB, ("sdma_addr (%qx, 0x%x) called\n",
				 (gm_u64_t) host_address, port));

#if !GM_ENABLE_VM
  return ((gm_dp_t) (host_address & ~gm.no_VM.clear) | gm.no_VM.set);
#else /* GM_ENABLE_VM */
  {
    gm_dp_t ret;

    ret = _dma_addr (host_address, port, gm.page_hash.bogus_sdma_ptr);
    GM_PRINT (GM_TRACE_LANAI_DMA, ("returning "GM_U64_TMPL"\n",
				   GM_U64_ARG (ret)));
    return ret;
  }
#endif /* GM_ENABLE_VM */
}

static inline gm_dp_t
rdma_addr (gm_up_t host_address, unsigned port)
{
#if !GM_ENABLE_VM
  return ((gm_dp_t) (host_address & ~gm.no_VM.clear) | gm.no_VM.set);
#else /* GM_ENABLE_VM */
  return _dma_addr (host_address, port, gm.page_hash.bogus_rdma_ptr);
#endif /* GM_ENABLE_VM */
}

/* This function performs a DMA of length LEN from the user virtual
   memory address PTR to the lanai address LAR.  Any DMA from a
   nonDMAable user virtual memory address causes the value 0xaaaaaaaa
   to be DMAd instead.

   NOTE: (ptr % GM_DMA_GRANULARITY) must equal (lar % GM_DMA_GRANULARITY). */

/* This cover macro allows for more informative debugging messages */
#define USER_SDMA(ptr, lar, len, port) do {				\
  GM_PRINT (GM_TRACE_LANAI_DMA,						\
	    ("USER_SDMA("GM_U64_TMPL",%p,%u,%d)\n",			\
	     GM_U64_ARG (ptr), lar, len, port));			\
  GM_DEBUG_CHAINED_DMA_ASSERT (NOTICED_NOT (SDMAING));			\
 									\
  /* ensure ptr and lar have the same alignment */			\
  gm_assert (GM_DMA_ALIGNED ((ptr) - (unsigned long) (lar)));		\
  gm_assert (GM_DMA_GRANULARITY <= len);				\
  gm_assert (len <= GM_MAX_DMA_CTR);					\
  gm_assert (GM_DMA_ALIGN (u32, len) > 0);				\
  _USER_SDMA (ptr, lar, len, port);					\
} while (0)

static inline gm_lp_t
_USER_SDMA (gm_up_t ptr, gm_lp_t lar, gm_u32_t len, gm_u32_t port_id)
{
  gm_lp_t ret = lar;

  GM_CALLED_WITH_ARGS
    ((GM_U64_TMPL",%p,%d,%d", GM_U64_ARG (ptr), lar, len, port_id));
  if (GM_DEBUG_PAGE_HASH)
    {
      gm_printf (GM_STR ("_USER_SDMA(%qx,%x,0x%x,0x%x) called.\n"),
		 (gm_u64_t) ptr, (unsigned int) lar, len, port_id);
    }
  LOG_DISPATCH (0, "_USER_SDMA() called");
  
#if GM_ENABLE_VM
  /* Set up a chained DMA for all the segments of user data.  We use
     "DMA channel 2" for the user DMA, since the other DMA channel
     must be used (via SDMA_ADDR) to perform the UVMA->DMAA
     translation (i.e.: the other channels descriptor queue is
     busy.)

     Care is taken to ensure that the DMA chain is always terminated, in
     case the FPGA is polling the chain. */
  
  {
    unsigned remaining;
    
    remaining = GM_PAGE_REMAINING (ptr);
	
    while (remaining < len)
      {
	LOG_DISPATCH (0, "building descriptor");
	gm_chained_dma_append (sdma_addr (ptr, port_id), lar, remaining,
			       GM_DMA_DESCRIPTOR_E2L);
	LOG_DISPATCH (0, "advancing");
	
	ptr += remaining;
	lar = (char *) lar + remaining;
	len -= remaining;
	remaining = GM_PAGE_LEN;
      }
    gm_chained_dma_append (sdma_addr (ptr, port_id), lar, len,
			   GM_DMA_DESCRIPTOR_E2L + GM_DMA_DESCRIPTOR_WAKE);
  }
#else /* !GM_ENABLE_VM */
  gm_chained_dma_append (sdma_addr (ptr, port_id), lar, len,
			 GM_DMA_DESCRIPTOR_E2L + GM_DMA_DESCRIPTOR_WAKE);
#endif /* !GM_ENABLE_VM */
  LOG_DISPATCH (0, "_USER_DMA() returning");
  GM_RETURN (ret);
}

/* This function peforms a DMA of length LEN from the LANai pointer
   LAR to the user virtual memory address PTR.  If any part of the DMA
   corresponds to nonDMAable user virtual memory addresses, then the
   data is dropped by DMAing it to a trash page in the host. */

/* This cover macro allows for more informative debugging messages */

#define USER_RDMA(_lar,_ptr,_len,_port) do {				\
  gm_lp_t USER_RDMA_lar;						\
  gm_up_t ptr;								\
  gm_u32_t len;								\
  gm_u32_t USER_RDMA_port;						\
									\
  USER_RDMA_lar = _lar;							\
  ptr = _ptr;								\
  len = _len;								\
  USER_RDMA_port = _port;						\
									\
  GM_DEBUG_CHAINED_DMA_ASSERT (NOTICED_NOT (RDMAING));			\
  gm_assert (USER_RDMA_lar);						\
  gm_assert (GM_DMA_ALIGNED (ptr));					\
  gm_assert (GM_DMA_ALIGNED (USER_RDMA_lar));				\
  gm_assert (len >= GM_DMA_GRANULARITY);				\
  _USER_RDMA (USER_RDMA_lar, ptr, len, USER_RDMA_port);			\
} while (0)

#define GM_DEBUG_USER_RDMA 0

static inline void
_USER_RDMA (gm_lp_t lar, gm_up_t ptr, gm_u32_t len, gm_u32_t port_id)
{
  GM_PRINT (GM_DEBUG_PAGE_HASH,
	    (GM_STR ("_USER_RDMA("GM_U64_TMPL",%x,0x%x,0x%x) called.\n"),
	     GM_U64_ARG (ptr), (unsigned int) lar, len, port_id));
  LOG_DISPATCH (0, "_USER_RDMA() called");
  
#if GM_ENABLE_VM
  /* Set up a chained DMA for all the segments of user data.  We use
     "DMA channel 2" for the user DMA, since the other DMA channel
     must be used (via RDMA_ADDR) to perform the UVMA->DMAA
     translation (i.e.: the other channels descriptor queue is
     busy.)

     Care is taken to ensure that the DMA chain is always terminated, in
     case the FPGA is polling the chain. */
  
  {
    unsigned remaining;

    remaining = GM_PAGE_REMAINING (ptr);
    while (remaining < len)
      {
	LOG_DISPATCH (0, "building descriptor");
	gm_chained_dma_append (rdma_addr (ptr, port_id), lar, remaining,
			       GM_DMA_DESCRIPTOR_L2E);

	LOG_DISPATCH (0, "advancing");
	ptr += remaining;
	lar = (char *) lar + remaining;
	len -= remaining;
	remaining = GM_PAGE_LEN;
      }
    LOG_DISPATCH (0, "terminating DMA chain");
    gm_chained_dma_append (sdma_addr (ptr, port_id), lar, len,
			   GM_DMA_DESCRIPTOR_L2E + GM_DMA_DESCRIPTOR_WAKE);
  }
#else /* !GM_ENABLE_VM */
  gm_chained_dma_append (sdma_addr (ptr, port_id), lar, len,
			 GM_DMA_DESCRIPTOR_L2E + GM_DMA_DESCRIPTOR_WAKE);
#endif /* !GM_ENABLE_VM */
  LOG_DISPATCH (0, "_USER_DMA() returning");
}

#endif _gm_user_dma_h_
