/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997, 1998, 1999, 2000 by Myricom, Inc.           *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: glenn@myri.com */

/**
   GM provides reliable ordered delivery between pairs of hosts on a Myrinet
   with two nonpreemptive priority levels.

   ---

   The driving force behind the design of this MCP are the following
   observations:

   [NOTE: The following is based on the original LANai4 dispatch, but
   remains qualitatively correct, although some of the numbers are out
   of date for the newer implementation.  --Glenn]

   (1) The LANai is composed of 4 event-driven interfaces: The sdma, send,
   recv, and rdma interfaces.
   (2) The MCP is responsible for setting up transactions on each interface,
   but can do so only when certain conditions are satisfied (e.g.: The
   send interface is idle and data is available to send.).  These events
   will be called "start events".
   (3) The MCP must be able to detect the completion of transactions on each
   interface.  These events will be called "finish events".

   It turns out that only 14 bits of information are needed to specify
   whether a transaction can be started or concluded.  6 of the bits
   are in the bottom byte of ISR and are maintained by the LANai
   hardware.  The remaining 8 bits can be stored in a general purpose
   register "state" and a mask can be stored in IMR such that
   "ISR&IMR|state" packs the 14 state bits into the bottom of a
   register.  Consequently, a 16K gm_event_index table can be constructed
   such that gm_event_index[ISR&IMR|state] returns the index of the
   highest-priority pending event.  Also, since "state" is a register,
   we can set or clear any set of state bits using a single
   or-immediate or and-immediate instruction.

   We can also construct a small "gm.handler" table to convert the event
   index into a pointer to the code to handle the event.  While we
   could have had the first table perform the full conversion, this
   solution takes about 1/4 the memory and allows the handler for any
   event to be changed dynamically in just 2 instructions.  Also the
   double-table solution takes only 2 more clock cycles per dispatch
   than the single-table solution.

   The result of all this dispatch trickery is that dispatching can
   contributes a best-case of only 35 instructions of latency to the
   best-case one-way message latency while providing a flexible and
   efficient state-transition mechanism for the MCP.

  ---

  WARNING: The following information is somewhat out of date,
  particularly the info about queuing.

  Conceptually, the GM MCP is structured as depicted in the following
  diagram.  In the diagram, rounded rectangles represent the state
  machines, which execute concurrently, the canonical queue symbol
  ("]O") represents queue data structures, and sharp rectangles represent
  other data structures.  The capacity of each queue is specified, with
  "R" being the number for receive tokens and "S" being the number of send
  tokens.


				    +-------+
				    |Pending|
				    |(n)ack |
		Network		    | table |	      Network
		   |		    +-------+		 ^
		   V	 received     ^	   |		 |	+-----+
	       .-------. (n)acks      |	   |	     .-------.	|addr |
	       | Recv  |--------------+--. `-------->| Send  |<-|table|
	       `-------'	      |	 |  ,--------`-------'	+-----+
		|      ^	      |	 |  |	      |	     ^
		|      |	      |	 |  |	      |	     |
		V     /^\ Free	      |	 |  |	      V	    /^\ Send chunks
	      | 2 |   >-< recv	      |	 |  |	Free| 2 |   >-<
	       >-<   | 2 |chunks      |	 |  |	send >-<   | 2 |
  recv chunks  \_/     ^	      |	 |  |  chunks\_/     ^
		|      |	      |	 |  |	      |	     |
		V      |   generated  |	 |  |	      |	     |
	       .--------.   (n)acks   |	 |  |	      |	     |
	       |  RDMA	|-------------'	 |  |	      |	     |
	       `--------'		 |  |	      |	     |
		|      ^		 |  |	      |	     |
		|      |		 |  |	      |	     |
		|     /^\ Free recv	 |  |	      |	     |
		|     >-< token lists	 |  |	      |	     |
		|    | R |array		 |  |	      |	     |
		|      ^		 |  |	      |	     |
		|      |		 |  |	      |	     |
		|      ^                 |  |         |      |
		|			 |  |	      |	     |
		|			 V  V	      V	     |
		|			+----+	     .--------.
		| ,---------------------|sent|------>|	SDMA  |
		| |			|list|	     `--------'
		V V			|arry|		     ^
	       |R+S|			+----+		     |
		>-<					     |
recv tokens	\_/ (---Host/LANai communication queues---) /^\ Send tokens
		 |					    >-<
		 |					   |R+S|
		 |					     ^
		 |					     |
		 V					     ^

  The GM MCP is comprised of four state machines: SDMA, Send, Recv,
  and RDMA.  The responsibilities of each state machines are as
  follows:

    SDMA:

    Remove message descriptors from the send message list, and check the
    message type.  Recv tokens are stored in the appropriate "free recv
    token list".  All other tokens (send tokens) are handled by
    SDMAing the message to be sent into pieces and placing the data in
    "send chunks", usually by DMAing it, as send chunks become available.

    Send:

    Send all enqueued chunks, prepending the correct route before the
    first chunk for a message, and marking the end of the message with a
    tail flit.	After sending a message, place the token for the sent
    message in the sent list for the corresponding connection.

    Also, send any pending ack or nack messages as indicated by the
    pending (n)ack table.

    Recv:

    Break received messages into chunks and pass them to the RDMA
    engine if the message header is valid.  Messages with invalid
    headers are discarded. Intercept ack and nack messages, using the
    information in acks and nacks to free the appropriate token(s)
    from the "sent list array" into the recv token queue, marking the
    token as a send token, and using the information in nacks to
    prepend the appropriate tokens from the sent list to the send
    message list.

    RDMA:

    Check sequence numbers of received messages and check the
    availability of a receive message buffer of the correct size in the
    second level (as indicated by the free recv message array) and drop
    and nack erroneous messages as appropriate (messages that have bad
    CRCs or are undeliverable because there is no recv token of the
    appropriate size are nacked).  DMA chunks into the approriate
    location in the receive message buffer indicated by the recv token,
    and if the message CRC checks out, pass the message to the recvd
    messages list.

  All elements in this system require only bounded pre-allocated
  storage.  This may not be obvious for the pending nack table or the
  sent list array.  The "pending (n)ack table" records only a single
  sequence number and an ack/nack indicator for each connection.  The
  send message list, sent list array, and free send message list form a
  loop in which message descriptors are conserved (as do the other
  three queues/free_list pairs.)

  Only a bounded number of connections are supported.  But since the
  state is small and only one connection is required per host on the
  network, a .5MB LANai board should be able to support O(1000)-node
  networks.  One caveat: Only network diameters <= 16 are supported.

  ---

  Acks and Nacks are worth a little explaining.  GM uses a form of the
  "go back N" protocol with NACKs (see Telecommunication Networks:
  Protocols, Modeling and Analysis by Mischa Schwartz, Addison-Wesley,
  ISBN 0-201-16423-X). "Go back N" is the preferred protocol when
  software overhead, rather than network capacity, is the network's
  limiting factor, because it wastes network bandwidth during error
  recovery to reduce software overhead relative to other protocols.

  For each connection, when each message is sent for the first time, it
  is given a successive sequence number.  When the message is received,
  the receiver checks the sequence number received against the
  expected sequence number (seqno).  If there is a discrepency, the
  receiver nacks the expected seqno, indicating that the sender should
  retransmit the expected message and all following messages and
  implicitly acknowledeging all earlier messages.  All out-of sequence
  messages are dropped. If the seqno of a received message agrees with
  the expected value, the seqno is acked.

  Upon receiving an ack OR nack, the sender (i.e.: the recipient of
  the (n)ack) frees all messages with lower sequence numbers.  Upon
  receiving a nack, the receiver prepends to the send queue all
  messages with sequence numbers greater than or equal to the nacked
  sequence number, rewinding the send stream to the message missed by
  the receiver (the "go back N" part of the protocol).

  The ack and nack messages themselves are not delivered reliably, and
  consequenly timers must be used to force retransmission of
  unacknowledged messages if the ack or nack is lost.  In many reliable
  communication systems, a timer is set for each transmitted message.
  In GM, a timer goes of periodically, at which time all
  unacknowledged messages are queued for retransmission.	 Depending
  upon the length of the timeout, this system either wastes bandwidth
  retransmitting messages unneccesarily or recovers from errors very
  slowly.  However, this system works efficiently on Myrinet, where
  errors are extremely rare, bandwidth is plenty, and latency is
  small.  */

#include <stdio.h>
#include "gm_bootstrap.h"
#include "gm_chained_dma.h"
#include "gm_cpp.h"
#include "gm_connection.h"
#include "gm_debug.h"
#include "gm_debug_blocking.h"
#include "gm_debug_lanai_tlb.h"
#include "gm_debug_node_id.h"
#include "gm_debug_send_tokens.h"
#include "gm_dma_perf.h"
#include "gm_enable_crc32.h"
#include "gm_enable_crc32.h"
#include "gm_enable_datagrams.h"
#include "gm_enable_debug_counters.h"
#include "gm_enable_directed_send.h"
#include "gm_enable_error_counters.h"
#include "gm_enable_ethernet.h"
#include "gm_enable_fast_small_send.h"
#include "gm_enable_mcp_send_length_check.h"
#include "gm_enable_packet_counters.h"
#include "gm_enable_paranoid_interrupts.h"
#include "gm_enable_pio_sends.h"
#include "gm_enable_security.h"
#include "gm_enable_shortcut.h"
#include "gm_enable_trace.h"
#include "gm_error_handlers.h"
#include "gm_heartbeat.h"
#include "gm_gcc_version.h"
#include "gm_lanai_def.h"
#include "gm_port.h"
#include "gm_pte.h"
#include "gm_send_tokens.h"
#include "gm_state.h"
#include "gm_trace.h"
#include "gm_types.h"
#include "gmcp.h"
#if GM_ENABLE_VM
#include "gm_page_hash.h"
#endif

/****************************************************************
 * Global variables
 ****************************************************************/

/* HACK: Reserve space for the stack, and tell crt0.s where the top of
   the stack is. */

gm_u32_t gm_stack[GM_STACK_WORDS];
void *gm_top_of_stack = (char *) gm_stack + sizeof (gm_stack);

/* CSPI hack */
#if L5
void *LAR = 0;
void *EAR = 0;
int DMA_DIR = 0;
#endif

void
gm_abort ()
{
  gm_always_assert (0);
}

#if GM_MIN_SUPPORTED_SRAM <= 256
void
gm_bcopy (const void *_from, void *_to, gm_size_t len)
{
  register char *from, *to;
  register unsigned long i;

  from = (char *) _from;
  to = (char *) _to;

  for (i = len; i; i--)
    {
      *to++ = *from++;
    }
}

void
gm_reverse_bcopy (void *_from, void *_to, unsigned long len)
{
  register char *from, *to;
  register unsigned long i;

  from = (char *) (((char *) _from) + len);
  to = (char *) (((char *) _to) + len);

  for (i = len; i; i--)
    {
      *to-- = *from--;
    }
}


#else


/* pipelined Duff's Device bcopy implementation with 2 byte read-ahead. */

#define BCOPY3() *t++ = a; a = *f++; *t++ = b; b = *f++; *t++ = c; c = *f++;

void
gm_bcopy (const void *from, void *to, gm_size_t len)
{
  const char *f;
  char *t, *limit;
  char a, b, c;

  f = from;
  t = to;
  limit = t + len;

  a = *f++;
  b = *f++;
  c = *f++;
  switch (len & 15)
    {
    case 0:
      while (t != limit)
	{
	  BCOPY3 ();
    case 13:
	  BCOPY3 ();
    case 10:
	  BCOPY3 ();
    case 7:
	  BCOPY3 ();
    case 4:
	  BCOPY3 ();
    case 1:
	  if (t + 1 == limit)
	    {
	      *t++ = a;
	      return;
	    }
	  BCOPY3 ();
    case 14:
	  BCOPY3 ();
    case 11:
	  BCOPY3 ();
    case 8:
	  BCOPY3 ();
    case 5:
	  BCOPY3 ();
    case 2:
	  if (t + 2 == limit)
	    {
	      *t++ = a;
	      *t++ = b;
	      return;
	    }
	  BCOPY3 ();
    case 15:
	  BCOPY3 ();
    case 12:
	  BCOPY3 ();
    case 9:
	  BCOPY3 ();
    case 6:
	  BCOPY3 ();
    case 3:
	  BCOPY3 ();
	}
    }
}

#undef BCOPY3

#define BCOPY3() *--t = a; a = *--f; *--t = b; b = *--f; *--t = c; c = *--f;

/* Just like bcopy, by copy the bytes in reverse order */

void
gm_reverse_bcopy (void *from, void *to, unsigned long len)
{
  char *f, *t, *limit;
  char a, b, c;

  f = (char *) from + len;
  t = (char *) to + len;
  limit = (char *) to;

  a = *--f;
  b = *--f;
  c = *--f;
  switch (len & 15)
    {
    case 0:
      while (t != limit)
	{
	  BCOPY3 ();
    case 13:
	  BCOPY3 ();
    case 10:
	  BCOPY3 ();
    case 7:
	  BCOPY3 ();
    case 4:
	  BCOPY3 ();
    case 1:
	  if (t - 1 == limit)
	    {
	      *--t = a;
	      return;
	    }
	  BCOPY3 ();
    case 14:
	  BCOPY3 ();
    case 11:
	  BCOPY3 ();
    case 8:
	  BCOPY3 ();
    case 5:
	  BCOPY3 ();
    case 2:
	  if (t - 2 == limit)
	    {
	      *--t = a;
	      *--t = b;
	      return;
	    }
	  BCOPY3 ();
    case 15:
	  BCOPY3 ();
    case 12:
	  BCOPY3 ();
    case 9:
	  BCOPY3 ();
    case 6:
	  BCOPY3 ();
    case 3:
	  BCOPY3 ();
	}
    }
}

#undef BCOPY3

#endif

/****************************************************************
 * Dispatch macros
 ****************************************************************/

/****************
 * dispatch debug support functions
 ****************
 These are are inline functions (instead of macros) so we can better
 tell what assertion fails from the file and line number reported in
 the GM_FAILED_ASSERTION_INTERRUPT. */

/* verify that the state is not too large for the state table. */


void gm_dispatch (void);
int main (void);

/* instrumenting */
#if GM_ENABLE_TRACE
 /* loic says: we are wasting 3 times GM_TRACEBUFSIZE to ensure
    gtracebuf is aligned on a 2*GM_TRACEBUFSIZE boundary, if is too
    much, another strategy si to use a base pointer and an index, then
    we need 2 global registers, and probably two more instructions per
    instrumentation point, anothe rpossibility is to specify somehow
    to the linker the right alignement.

    glenn says: The linker cannot align on coarser than an 8 byte
    boundary. */
gm_l_trace_t gtracebufarea[GM_LANAI_NUMTRACE * 4];
gm_l_trace_t *gtracebuf;
#endif

#include "gm_dispatch.h"

/************************************************************************
 * Miscellaneous functions
 ************************************************************************/

#include "gm_rand.h"

void
gm_bzero (void *s_pv, gm_size_t c)
{
#if GM_MIN_SUPPORTED_SRAM <= 256
  register char *cptr;
  register gm_size_t i;

  cptr = (char *) s_pv;
  for (i = c; i; i--)
    {
      *cptr++ = 0;
    }
#else /* GM_MIN_SUPPORTED_SRAM > 256 */
  gm_s32_t *top, *bottom;
  gm_s8_t *s = (gm_s8_t *) s_pv;

  gm_assert (s_pv);

  if (c <= 3)
    {
      gm_size_t i;

      for (i = 0; i < c; i++)
	s[i] = 0;
      return;
    }

  /* Clear the top and bottom three bytes of the region. */

  s[0] = s[1] = s[2] = s[c - 1] = s[c - 2] = s[c - 3] = 0;

  /* Clear all the ints (4 bytes) within the string in a fast unrolled loop. */

  bottom = (gm_s32_t *) ((unsigned) (s + 3) & ~3);
  top = (gm_s32_t *) ((unsigned) (s + c) & ~3);
  if (bottom >= top)
    return;
  switch ((top - bottom) & 0xf)
    {
    case 0:
      while (bottom < top)
	{
	  *bottom++ = 0;
    case 0xf:
	  *bottom++ = 0;
    case 0xe:
	  *bottom++ = 0;
    case 0xd:
	  *bottom++ = 0;
    case 0xc:
	  *bottom++ = 0;
    case 0xb:
	  *bottom++ = 0;
    case 0xa:
	  *bottom++ = 0;
    case 0x9:
	  *bottom++ = 0;
    case 0x8:
	  *bottom++ = 0;
    case 0x7:
	  *bottom++ = 0;
    case 0x6:
	  *bottom++ = 0;
    case 0x5:
	  *bottom++ = 0;
    case 0x4:
	  *bottom++ = 0;
    case 0x3:
	  *bottom++ = 0;
    case 0x2:
	  *bottom++ = 0;
    case 0x1:
	  *bottom++ = 0;
	}
    }
  return;
#endif /* GM_MIN_SUPPORTED_SRAM > 256 */
}

/* Compare 32-bit sequence numbers. Return 0 if equal, negative if
   a<b, and positive if a>b, but compensating for wraparound. */

#define SEQ_CMP32(a, b) ((gm_s32_t) (a) - (gm_s32_t) (b))

/* Compare 16-bit sequence numbers. Return 0 if equal, negative if
   a<b, and positive if a>b, but compensating for wraparound */

#define SEQ_CMP16(a, b) (((gm_s32_t) (a) - (gm_s32_t) (b)) << 16)

/******************************
 * Host <-> LANai communication
 ******************************/

/****************
 * putstring
 ****************/

void
gm_putstring (char *my_str)
{
  prepare_to_interrupt ("gm_putstring ");
  gm.interrupt.print.string = my_str;
  gm_interrupt (GM_PRINT_INTERRUPT);
}

#define GM_DEBUG_HANDLE_SEND_ERROR 0

void
handle_send_error (gm_subport_t * sp, gm_status_t errno, int dest_id)
{
  gm_send_token_t *st;
  unsigned int sp__id, port__open;
  gm_port_protected_lanai_side_t *port;

  gm_assert (sp->first_send_token);

  /* pre */ sp__id = sp->id;
  /* pre */ st = sp->first_send_token;
  ;
  port = &gm_port[GM_SUBPORT_PORT (sp__id)];
  ;
  ;
  port__open = port->open;
  ;
  ;
#if GM_DEBUG || GM_DEBUG_HANDLE_SEND_ERROR
  gm_printf ("handle_send_error: sending subport=%d   errno=%d  dest_GMid=%d\n",
			sp__id, errno, dest_id);
  switch (errno)
   {
     case GM_SEND_TIMED_OUT:
        gm_printf ("handle_send_error: errno=%d  ==> '%s'\n",errno, "Send timed out - no ACK");
        break;
     case GM_SEND_REJECTED:
        gm_printf ("handle_send_error: errno=%d  ==> '%s'\n",errno, "Send was rejected");
        break;
     case GM_SEND_TARGET_PORT_CLOSED:
        gm_printf ("handle_send_error: errno=%d  ==> '%s'\n",errno, "Target port was closed");
        break;
     case GM_SEND_TARGET_NODE_UNREACHABLE:
        gm_printf ("handle_send_error: errno=%d  ==> '%s'\n",errno, "Target node unreachable");
        break;
     default:
        gm_printf ("handle_send_error: errno=%d  ==> '%s'\n",errno, "???");
       break;
   }
  gm_log_call_trace ();
  fflush (stdout);
#endif /* GM_DEBUG */

  if (port__open)
    {
      gm_send_token_t *next;

      /* report the failed send to the user. */

      next = st->common.next;
      LOG_DISPATCH (0, "disabling subport");
      sp->disabled = 2;
      if (next)
	{
	  remove_first_send_token_from_send_queue (st, sp__id);
	}
      else
	{
	  /* HACK: defer removal of the subport so it can be marked
	     disabled. */

	  sp->first_send_token = 0;
	  sp->last_send_token
	    = (gm_send_token_t *) ((char *) &sp->first_send_token
				   - GM_OFFSETOF (gm_send_token_t,
						  common.next));
	}

      pass_sent_token_to_port_and_free (st, port, 1, errno);
    }
  else
    {
      /* free all send tokens and the subport */

      gm_assert (st == sp->first_send_token);
      gm_assert (st);
      do
	{
	  remove_first_send_token_from_send_queue (st, sp__id);
	  pass_sent_token_to_port_and_free (st, port, 1, errno);
	}
      while ((st = sp->first_send_token) != 0);
    }
}

/**********************************************************************/

/**********************************************************************/

static void
gm_initialize (void)
{
  unsigned int h, i, j;
  static int initialize_called;

  /* Make sure the LANai is not re-initialized.  This can happen if,
     for example, the LANai is reset or accidentally branches to
     memory location 0. */

  /* Macros for temporary salting of location 0 with the line number */
#define GM_ENABLE_SALT_LINENO 0
#if GM_ENABLE_SALT_LINENO
#define GM_SALT_LINENO \
  *(gm_u32_t *) 0 = (0x43210000 + __LINE__);
#else
#define GM_SALT_LINENO /* nothing */
#endif

  *(gm_u32_t *) 0 = 0x12340000;
  while (initialize_called)
    ;
  initialize_called = 1;
  *(gm_u32_t *) 0 = 0x12341111;

  /****************
   * Pre-inialization
   ****************/

  /* The host cleared all of the LANai SRAM before loading the MCP, so
     it is safe to assume that all uninitialized variables are cleared. */


  /* Initialize ubiquitous globals */

#if GM_ENABLE_MCP_GLOBAL_REGS
  gmptr = &_gmp.globals;
  gm_connection = gm.connection;
  gm_port = gm.port;
  /*  GM_PORT = gm._PORT; */
#endif

  /****************
   * Hardware Initialization
   ****************/

  gm_init_lanai_hardware ();

  /****************
   * CPU Register setup
   ****************/

  /********************* CPU Register setup *********************/

  /* Clear the globals that *might* be in hard registers */

  GM_STATE = 0;
  GM_PORT = 0;			/* reinitialized below, once the SRAM
				   length is known. */


  *(gm_u32_t *) 0 = 0x12342222;

  /****************
   * State Machine initialization
   ****************/

  /* Set IMR to participate in DISPATCH( ) */

  set_IMR (IMR_INIT_VAL);

  /* Initialize State */

  NOTICE (FREE_RECV_CHUNK + FREE_SEND_CHUNK);

  NOTICE_NO (RDMA_PENDING
	     + RDMAING
	     + RECEIVING
	     + SDMAING + SEND_PENDING + ACK_PENDING + SENDING + SDMA_PENDING
	     + SDMA_INT_BIT + RDMA_INT_BIT);

  /* Set up gm_event_index table to convert between state and event
     indices.  I.e.: the_index=gm_event_index[state]. */

  for (i = 0; i < (BIGGEST_STATE_BIT << 1); i++)
    {
      /* Highest priority so queues will be rewound when needed. */

      if (i & TIME_INT_BIT)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (TIMER_EVENT, i);
	  continue;
	}

      /* Translate DMA_INT_BITs into SDMA_INT_BITs and RDMA_INT_BITS.
	 This must be done with higher priority than the handlers that
	 depend on these bits. */

      if ((~i & SDMAING
	   && ~i & SDMA_INT_BIT)
	  || (~i & RDMAING
	      && ~i & RDMA_INT_BIT))
	{
	  /* Catch these DMA states, which should never happen. */
	  GM_SET_EVENT_INDEX_FOR_STATE (DMA_PANIC_EVENT, i);
	  continue;
	}
      
      if (i & DMA_INT_BIT
	  && (i & SDMAING
	      && ~i & SDMA_INT_BIT)
	  && (i & RDMAING
	      && ~i & RDMA_INT_BIT)
	  )
	{
	  /* Got a DMA_INT_BIT when multiple DMAs are in progress.
	     Determine which DMA completed, or if both completed. */
	  GM_SET_EVENT_INDEX_FOR_STATE (DMA_INT_EVENT, i);
	  continue;
	}
      if (i & DMA_INT_BIT
	  && (i & SDMAING
	      && ~i & SDMA_INT_BIT))
	{
	  /* Only an SDMA is in progress. */
	  GM_SET_EVENT_INDEX_FOR_STATE (SDMA_INT_EVENT, i);
	  continue;
	}
      if (i & DMA_INT_BIT
	  && (i & RDMAING
	      && ~i & RDMA_INT_BIT))
	{
	  /* Only an RDMA is in progress. */
	  GM_SET_EVENT_INDEX_FOR_STATE (RDMA_INT_EVENT, i);
	  continue;
	}
      if (i & DMA_INT_BIT)
	{
	  /* Got a DMA interrupt when no DMA was in progress. */
	  GM_SET_EVENT_INDEX_FOR_STATE (DMA_PANIC_EVENT, i);
	  continue;
	}

      /* RDMA completion */
      
      if (1
	  && i & RDMA_INT_BIT
	  && i & RDMAING)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (FINISH_RDMA_EVENT, i);
	  continue;
	}

      /* receive events */

      if (i & RECV_INT_BIT && i & RECEIVING)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (FINISH_RECV_PACKET_EVENT, i);
	  continue;
	}
      if (i & BUFF_INT_BIT && i & RECEIVING)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (RECV_BUFFER_OVERFLOW_EVENT, i);
	  continue;
	}
      if (~i & RECEIVING && i & FREE_RECV_CHUNK)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (START_RECV_PACKET_EVENT, i);
	  continue;
	}

      /* send events */

      if (i & SEND_INT_BIT && i & SENDING)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (FINISH_SEND_EVENT, i);
	  continue;
	}
      if (i & SEND_INT_BIT && ~i & SENDING && i & ACK_PENDING)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (SEND_ACK_EVENT, i);
	  continue;
	}
      if (i & SEND_INT_BIT && ~i & SENDING && i & SEND_PENDING)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (START_SEND_EVENT, i);
	  continue;
	}

      /* SDMA completion events */

      if (1
	  && i & SDMA_INT_BIT
	  && i & SDMAING)
	{
	  GM_SET_EVENT_INDEX_FOR_STATE (FINISH_SDMA_EVENT, i);
	  continue;
	}

      /* Interlocked SDMA/RDMA state machine fairness events */

      {
	int bits;

	bits = 0;

	if (1
	    && i & SDMA_INT_BIT
	    && ~i & SDMAING
	    && i & FREE_SEND_CHUNK
	    && i & SDMA_PENDING)
	  bits |= SDMA_PENDING;

	if (1
	    && i & RDMA_INT_BIT
	    && ~i & RDMAING
	    && i & RDMA_PENDING)
	  bits |= RDMA_PENDING;

	switch (bits)
	  {
	  case (SDMA_PENDING + RDMA_PENDING):
	    GM_SET_EVENT_INDEX_FOR_STATE (FAIR_SDMA_RDMA_EVENT, i);
	    continue;
	  case (SDMA_PENDING):
	    GM_SET_EVENT_INDEX_FOR_STATE (FAIR_SDMA_EVENT, i);
	    continue;
	  case (RDMA_PENDING):
	    GM_SET_EVENT_INDEX_FOR_STATE (FAIR_RDMA_EVENT, i);
	    continue;
	  case (0):
	    GM_SET_EVENT_INDEX_FOR_STATE (POLL_EVENT, i);
	    continue;
	  default:
	    GM_ABORT ();
	  }
      }
    }
  gm.event_index_table_crc = gm_crc (_gmp.event_index,
				     sizeof (_gmp.event_index));

  /********************************
   * Host synchronization...
   *
   * ... is performed as follows:
   * LANai sets SMP = &gm so the host can find the globals.
   * Host set gm.sram_length so the LANai can compute max_node_id.
   * Host sets important values in LANai SRAM, including some that
   *    require max_node_id to create, and then clears SMP.
   ********************************/

  /* Tell host where to find the lanai globals by putting a pointer
     in the SMP because it is
     (a) cleared during LANai reset
     (b) well-known to the host
     (c) guaranteed not to change until the host finds it.

     We also tag the beginning and end of the globals to allow the host
     to catch inconsistencies in the MCP size. */

  gm.magic = gm.end_magic = 0xcafebabe;
  gm.length = sizeof (gm);

  GM_STBAR ();

  SMP = &gm;

  GM_STBAR ();

  /* HACK to make sure "ZZZ" is not flashed unless host sleeps too long. */

  heartbeat ();

  /****************
   * High memory setup
   *
   * This is done early so the host can wait for max_node_id, which it
   * needs to initialize the host-side ethernet addr and hostname
   * tables.
   ****************/

  *(gm_u32_t *) 0 = 0x12343333;

  /* Wait for the host to set lanai_sram_size */

  while (gm.sram_length == 0)
    {
      GM_STBAR ();
    }

  gm_assert (GM_PORT_MAPPED_SRAM_LEN
	     == sizeof (gm_port_unprotected_lanai_side_t));
  gm_assert (gm.sram_length);
  GM_PORT = ((gm_port_unprotected_lanai_side_t *) 
	     (gm.sram_length - GM_PAGE_HASH_PIECE_REF_TABLE_LEN
	      - (GM_NUM_PORTS * GM_PORT_MAPPED_SRAM_LEN)));

  /* Figure how much space to use for the connection array and how
     much for cached PTEs.  Also determine the location of the cached
     PTE table. */

  {
    unsigned long available_memory;
    unsigned long cached_pte_cnt;
    unsigned long connection_cnt;
    unsigned long const GM_MAX_CONNECTION_CNT = 
#if L4 | L5 
						511+1;
#else
						3000+1;
#endif
	
    gm_assert (GM_PORT);
    available_memory = (char *) GM_PORT - (char *) gm.connection;
    connection_cnt = available_memory / 2 / sizeof (gm_connection_t);
    if (connection_cnt > GM_MAX_CONNECTION_CNT)
      {
	connection_cnt = GM_MAX_CONNECTION_CNT;
      }
    available_memory -= connection_cnt * sizeof (gm_connection_t);
    cached_pte_cnt = available_memory / sizeof (gm_cached_pte_t) - 1;
    if (!GM_POWER_OF_TWO (cached_pte_cnt))
      {
	cached_pte_cnt = 1 << (gm_log2_roundup (cached_pte_cnt) - 1);
      }
    available_memory -= (cached_pte_cnt + 1) * sizeof (gm_cached_pte_t);
    connection_cnt += available_memory / sizeof (gm_connection_t);
    if (connection_cnt > GM_MAX_CONNECTION_CNT)
      {
	connection_cnt = GM_MAX_CONNECTION_CNT;
      }

    gm.max_node_id = connection_cnt - 1;
    gm.page_hash.cache.entry = (void *) &gm.connection[connection_cnt];
    GM_MAX_PAGE_HASH_CACHE_INDEX = cached_pte_cnt - 1;
  }

  /****************
   * Get some important values from the host
   ****************/

  /* Wait for host initialization to complete.  The host sets the
     DMA_STS, BURST, and VERSION registers, as needed while we wait.  This
     takes several milliseconds. */

  while (SMP)
    {
      gm.while_waiting++;
      gm_flash_while_waiting ("ZZZ\n");
    }

  /* Verify the host set all it should have set. */

  gm_assert (gm.page_hash.bogus_sdma_ptr);
  gm_assert (gm.page_hash.bogus_rdma_ptr);
  gm_assert (gm.sram_length);

  *(gm_u32_t *) 0 = 0x12344444;
  
  /****************
   * global initialization
   ****************/

  /* gm.magic set above during host synchronization */
  gm._PORT = GM_PORT;
  /* gm.dma_descriptor = {0}; */
  /* gm.flush_dma_descriptor = {0}; */
#if L4 | L5
  /* nothing to do */
#elif L6 | L7 | L8 | L9
  gm.flush_dma_descriptor.next_with_flags = (gm_lp_t) DMA_TERMINAL;
  gm.user_dma_descriptor[0].next_with_flags = (gm_lp_t) DMA_TERMINAL;
#else
#error Unrecognized CPU.
#endif
  gm.flush_dma_descriptor.lanai_addr = gm.dummy_flush_target;
  gm.flush_dma_descriptor.len = 2 * GM_SPARC_STREAMING_FLUSH_LEN;
  gm_always_assert (sizeof (gm.dummy_flush_target) >= gm.flush_dma_descriptor.len);

#if L4 | L5
  /* nothing to do */
#elif L6 | L7 | L8 | L9
  gm.dma_descriptor.next_with_flags = (gm_lp_t) DMA_TERMINAL;
#else
#error Unrecognized CPU.
#endif

  gm.event_index_table = gm_event_index;	/* for debugging */
  /* gm.this_node_id = 0 */
  /* gm.port_to_wake = 0; */
  gm_assert (gm.page_hash.bogus_sdma_ptr);	/* set by host */
  gm_assert (gm.page_hash.bogus_rdma_ptr);	/* set by host */
#if GM_ENABLE_VM
  gm_init_page_table ();	/* gm.page_hash.cache */
#endif /* GM_ENABLE_VM */
  /* gm.first_port_with_sent_packets = 0; */

  gm.free_subports = &gm._subport[0];
  for (j = 0; j < GM_NUM_SUBPORTS - 1; j++)
    gm._subport[j].next = &gm._subport[j + 1];

  gm.timeout_time = RTC;
  gm_assert (gm.sram_length);	/* set by host */

  /* gm._state initialized above */
  gm.free_recv_chunk_cnt = 2;
  /* gm.current_rdma_port = 0 */
  /* gm.registered_raw_recv_port = 0 */
  gm.free_send_chunk_cnt = 2;
  /* gm.first_connection_to_ack = 0; */

  /* gm.send_chunk = {0}; */
#if GM_CPU_lanai
  gm.send_chunk[0].send_list = gm.send_chunk[0].packet.as_gm.payload;
  gm.send_chunk[1].send_list = gm.send_chunk[1].packet.as_gm.payload;
  gm.send_chunk[0].send_list_end = gm.send_chunk[0].packet.as_gm.payload;
  gm.send_chunk[1].send_list_end = gm.send_chunk[1].packet.as_gm.payload;
#endif
  gm.send_chunk[0].packet.as_gm.header.type = GM_PACKET_TYPE;
  gm.send_chunk[1].packet.as_gm.header.type = GM_PACKET_TYPE;
  gm.send_chunk[0].cafebabe = 0xcafebabe;
  gm.send_chunk[1].cafebabe = 0xcafebabe;
  gm.ethernet.send.chunk[0].cafebabe = 0xcafebabe;
  gm.ethernet.send.chunk[1].cafebabe = 0xcafebabe;
  /* gm.recv_chunk = {0}; */
  gm.recv_chunk[0].cafebabe = 0xcafebabe;
  gm.recv_chunk[1].cafebabe = 0xcafebabe;
  /* gm.recv_token_bin = {0}; */
  /* gm.first_active_connection = 0; */

  gm.free_send_records = &gm._send_record[0];
  for (j = 0; j < GM_NUM_SEND_RECORDS - 1; j++)
    gm._send_record[j].next = &gm._send_record[j + 1];

  /* gm._remaining_sdma_ctr = 0; */
  /* gm._remaining_rdma_ctr = 0; */
  /* gm.remaining_sdma_lar = 0; */
  /* gm.remaining_rdma_lar = 0; */
  /* gm.remaining_sdma_ear = 0; */
  /* gm.remaining_rdma_ear = 0; */

  /* gm.failed_send_dma_stage = {0}; */
  /* gm.failed_send_dma_stage.type = set dynamically */

  /* gm.report_dma_stage = {0}; */

  /* gm.nack_delay = 0 */
  gm.rand_seed = -1;
  /* gm.backlog_delay = 0 */
  /* gm.pause_rqst = 0 */
  /* gm.pause_ack = 0 */
  gm.port_to_close = -1;
  gm.port_to_open = -1;
  /* gm_failed_line = 0; */
  /* gm_failed_file = 0; */

/*
  gm.nack_cnt = 0;
  gm.nack_down_recv_cnt = 0;
  gm.nack_down_send_cnt = 0;
  gm.nack_reject_cnt = 0;
  gm.nack_received_cnt = 0;
  gm.nack_normal_cnt=0;
  gm.nack_send_nothing1_cnt=0;
  gm.nack_send_nothing2_cnt=0;
  gm.nack_send_open_connection_cnt=0;
  gm.nack_send_close_connection_cnt=0;
  gm.nack_receive_close_connection_cnt=0;
  gm.nack_ignore_close_connection_cnt=0;
  gm.handle_connection_reset_request_cnt=0;
  gm.nack_ignored_cnt = 0;
  gm.drop_cnt = 0;
  gm.resend_cnt = 0;
  gm.netrecv_cnt = 0;
  gm.too_small_cnt = 0;
  gm.bogus_header_cnt = 0;
  gm.out_of_sequence_cnt = 0;
*/
  /* gm._reserved_after_resent_cnt = 0 */
  /* gm.led = 0 */
  set_LED (0);

  for (i = 0; i < GM_NUM_PORTS; i++)
    gm_init_port (i);

  /* gm.finishing_rdma_for_port = 0; */
  /* gm.first_port_with_alarm = 0; */

  /* gm.record_log = 0; */
#if GM_LOG_DISPATCHES
  gm.log_slot = gm.log;
  gm.log_end = &gm.log[GM_LOG_LEN];
  gm.logtime_index = 0;
#endif
  /* gm.lzero = 0; */
  /* gm.log = {0} */
  gm.current_handler = "none";
  gm.current_handler_extra = "";
  /* gm.resume_after_halt = 0; */
  /* gm.volatile_zero = 0; */
  for (j = 0; j < 2; j++)
    for (i = 0; i < 128; i++)
      {
	/* gm.dispatch_cnt[i][j]=0; */
      }

  /* gm.hit_cnt = 0 */
  /* gm.miss_cnt = 0 */
  /* gm.pause_cnt = 0 */
  /* gm.hashed_token_cnt = 0 */

  /***
   * Begin mapping state initialization
   ***/

  /*** initialize the mapping scout message reply packet ***/
  /* gm.mapper_state.scout_reply.route_len = 0; */
  /* gm.mapper_state.scout_reply.in_send_queue = 0 */
  /* gm.mapper_state.scout_reply.route[*] = 0 */
  gm.mapper_state.scout_reply.packet.type = GM_MAPPING_PACKET_TYPE;
  gm.mapper_state.scout_reply.packet.subtype =
    GM_MAPPER_SCOUT_REPLY_PACKET_SUBTYPE;
  /* gm.mapper_state.scout_reply.packet.port = don't care */
  /* gm.mapper_state.scout_reply.packet.phase = don't care */
  /* gm.mapper_state.scout_reply.packet.address = initialized by host */
  gm.mapper_state.scout_reply.packet.option =
    GM_MAPPER_SCOUT_REPLY_PACKED_ROUTES_OPTION |
    GM_MAPPER_SCOUT_REPLY_LONG_HOSTNAME_OPTION;
  /*gm.mapper_state.scout_reply.packet.extendHostname = initialized by host */

  /*should the MCP initialized this? */
  gm.mapper_state.scout_reply.packet.node_type = 0;

  gm.mapper_state.scout_reply.send_token.common.type
    = GM_ST_MAPPER_SCOUT_REPLY;
  gm.mapper_state.scout_reply.send_token.common.sendable = 1;

  /*** initialize the mapping config message reply packet ***/
  /* gm.mapper_state.config_reply.route_len = 0; */
  /* gm.mapper_state.config_reply.in_send_queue = 0 */
  /* gm.mapper_state.config_reply.route[*] = 0 */
  gm.mapper_state.config_reply.packet.type = GM_MAPPING_PACKET_TYPE;
  gm.mapper_state.config_reply.packet.subtype =
    GM_MAPPER_CONFIG_REPLY_PACKET_SUBTYPE;
  /* gm.mapper_state.config_reply.packet.port = don't care */
  /* gm.mapper_state.config_reply.packet.phase = don't care */
  /* gm.mapper_state.config_reply.packet.address = initialized by host */
  /* gm.mapper_state.config_reply.packet.host_section = 0 */
  gm.mapper_state.config_reply.send_token.common.type
    = GM_ST_MAPPER_CONFIG_REPLY;
  gm.mapper_state.config_reply.send_token.common.sendable = 1;

  /***
   * End mapping state initialization
   ***/

  /* gm.ethernet = {0} */
  gm.ethernet.send.token.ethernet.sendable = 1;
  gm.ethernet.recv.token_slot = &gm.ethernet.recv.token[0];
  /* gm.trash = {0} */
  gm_assert (gm.max_node_id);	/* set by host */
  /* gm.end_magic set above during host synchronization */

  gm_chained_dma_initialize ();
  
  /********************* Connection initialization *********************/

  /* Initialize the connections, which fill the space between the end
     of LANai memory and the stack.  The number of connections is one
     more than gm.max_node_id, which is computed by the host. */

#if 0
  gm_printf ("%s:%d  gm.max_node_id = %d\n", __GM_FILE__, __LINE__,
	     gm.max_node_id);
  gm_printf ("GM_PORT = %p, gm.connection = %p -> %p\n",
	  (char *) GM_PORT, (char *) gm.connection,
	  (char *) &gm.connection[gm.max_node_id]);
  gm_printf ("%d\n", (gm_connection_t *) GM_PORT - gm.connection);
#endif

  gm_always_assert ((char *) &gm.connection[gm.max_node_id] <=
		    (char *) GM_PORT);

  for (h = 0; h <= gm.max_node_id; h++)
    {
      _gm_connection_init (&gm_connection[h]);
      set_LED (RTC >> 17);
    }

  *(gm_u32_t *) 0 = 0x12345555;

  /********************* ISR setup *********************/

  /* Setup the first receive */

  RMP = gm.recv_chunk[0].packet.as_bytes;
  set_RML (&gm.recv_chunk[0].end);
  NOTICE (RECEIVING);

#ifdef BAD
  /* Verify ISR was reset */

  gm_assert ((get_ISR () & IMR & ~TIME_INT_BIT) == 0);
#endif

  /* Set the SEND_INT_BIT */

  /* Harmlessly send an ack to this node, causing send_int_bit to be set. */

  SMP = &gm_connection[gm.this_node_id].ack_packet;
  set_SMLT (&gm_connection[gm.this_node_id].ack_packet + 1);
  while (!(get_ISR () & SEND_INT_BIT))
    set_LED (RTC >> 17);


  *(gm_u32_t *) 0 = 0x12346666;

  /* Set SDMA_INT_BIT and RDMA_INT_BIT to indicate that it is safe to
     start SDMAs and RDMAs. */

  NOTICE (SDMA_INT_BIT + RDMA_INT_BIT);

  /* Make sure the DMA interrupt bit was never set during reset */
  
  gm_assert ((get_ISR () & DMA_INT_BIT) == 0);

  /* Setup idle state of SDMA and RDMA to indicate that all DMAs are done. */
  
  gm.dma.fake_terminal_dma_descriptor.next_with_flags
    = (gm_lp_n_t) GM_DMA_DESCRIPTOR_TERMINAL;
  gm.sdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
  gm.rdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
  
  /*
   * Set the SDMA_INT_BIT and RDMA_INT_BIT by using the DMA engine.
   *
   * We take this opportunity to measure the performance of both
   * send and receive DMAs.
   */

  gm_assert (gm.page_hash.bogus_sdma_ptr);

  *(gm_u32_t *) 0 = 0x12346677;

  {
    int loop;
    unsigned int start;

    gm.e2l_time[0] = gm.l2e_time[0] = 0;
    for (loop = GM_DMA_PERF_NUM_PAGES; loop; loop--) 
      {
  
        start = RTC;
        {
          gm_s32_t remaining;
          for (remaining = GM_PAGE_LEN; remaining > 0; remaining -= GM_MTU)
            {
              GM_SALT_LINENO;
	      synchronous_SDMA
		(gm.page_hash.bogus_sdma_ptr,
		 gm.send_chunk[0].packet.as_bytes,
		 (remaining<=GM_MTU ? remaining : GM_MTU));
              GM_SALT_LINENO;
            }
        }
        gm.e2l_time[0] += (RTC - start);
        GM_SALT_LINENO;
      } /* for loop */

    GM_SALT_LINENO;
    LOG_DISPATCH (0, "bogus send dma ended...\n");
    GM_SALT_LINENO;
    gm_assert (*(gm_u32_t *)gm.send_chunk[0].packet.as_bytes == 0xaaaaaaaa);
    GM_SALT_LINENO;
    
    /* RDMA a page */
      
    GM_SALT_LINENO;
    gm_assert (gm.page_hash.bogus_rdma_ptr);
    GM_SALT_LINENO;
	
    for (loop = GM_DMA_PERF_NUM_PAGES; loop; loop--) 
      {
        start = RTC;
        {
          gm_s32_t remaining;
          for (remaining = GM_PAGE_LEN; remaining > 0; remaining -= GM_MTU)
            {
	      synchronous_RDMA
		(gm.send_chunk[0].packet.as_bytes,
		 gm.page_hash.bogus_rdma_ptr,
		 remaining<=GM_MTU ? remaining : GM_MTU);
  	      LOG_DISPATCH (0, "bogus recv dma started...\n");
            }
        }
        gm.l2e_time[0] += (RTC - start);
        LOG_DISPATCH (0, "bogus recv dma ended...\n");
      } /* for loop */

  }

#if GM_DMA_PERF_ALTERNATIVE_TIMING
  GM_SALT_LINENO;

  {
    int loop;
    volatile unsigned int start;

    for (i = 1; i < 4; i++)
    {
     /* First loop:  the "old" way.
      * Second loop: Repeat the "old" way for comparison.
      * Third loop:  Alternate the page used for each of the
      *              GM_DMA_PERF_NUM_PAGES passes.
      * Fourth loop: Alternate the page used for each DMA.
      * To be added if desired: Alternate send and receive DMA.
      */  
      gm.e2l_time[i] = gm.l2e_time[i] = 0;
      gm_assert (gm.page_hash.bogus_rdma_ptr);
      for (loop = GM_DMA_PERF_NUM_PAGES; loop; loop--)
      {
	{
	  register gm_s32_t remaining;
          register int      inner_count = 0;

	  start = RTC;
	  for (remaining = GM_PAGE_LEN; remaining > 0; remaining -= GM_MTU)
	  {
            if (((i == 2) && (loop % 2)) || ((i == 3) && (inner_count % 2)))
              {
	        synchronous_SDMA
		  (gm.page_hash.bogus_rdma_ptr,
		   gm.send_chunk[0].packet.as_bytes,
		   (remaining <= GM_MTU ? remaining : GM_MTU));
              }
            else
              {
	        synchronous_SDMA
		  (gm.page_hash.bogus_sdma_ptr,
		   gm.send_chunk[0].packet.as_bytes,
		   (remaining <= GM_MTU ? remaining : GM_MTU));
              }
            inner_count++;
	  }
	}
        gm.e2l_time[i] += (RTC - start);

	if (GM_DEBUG
	    && *(gm_u32_t *) gm.send_chunk[0].packet.as_bytes != 0xaaaaaaaa)
	  {
	    gm_hex_dump (gm.send_chunk[0].packet.as_bytes, 64);
	    GM_ABORT ();
	  }

	/* RDMA a page */

	{
	  register gm_s32_t remaining;
          register int      inner_count = 0;

	  start = RTC;
	  for (remaining = GM_PAGE_LEN; remaining > 0; remaining -= GM_MTU)
	  {
            if (((i == 2) && (loop % 2)) || ((i == 3) && (inner_count % 2)))
              {
	        synchronous_RDMA
		  (gm.send_chunk[0].packet.as_bytes,
		   gm.page_hash.bogus_sdma_ptr,
		   remaining <= GM_MTU ? remaining : GM_MTU);
              }
            else
	      {
	        synchronous_RDMA
		  (gm.send_chunk[0].packet.as_bytes,
		   gm.page_hash.bogus_rdma_ptr,
		   remaining <= GM_MTU ? remaining : GM_MTU);
	      }
            inner_count++;
	  }
	}
        gm.l2e_time[i] += (RTC - start);
      }				/* for loop */

     } /* for GM_DMA_PERF_NUM_PAGES loop */
  } /* for i */
#endif /* GM_DMA_PERF_ALTERNATIVE_TIMING */
  
  *(gm_u32_t *) 0 = 0x12346699;

  set_IT (1);

  gm.current_handler = "0x12347777";
  *(gm_u32_t *) 0 = 0x12347777;
  GM_STBAR ();

  LOG_DISPATCH (0, "gm_initialized\n");

#if GM_ENABLE_TRACE
  gtracebuf =
    (void *) (((unsigned) gtracebufarea + 2 * GM_TRACEBUFSIZE - 1) &
	      ~(2 * GM_TRACEBUFSIZE - 1));
  G_TRACEPTR = (unsigned) gtracebuf;
  gm.trace_index = 0;
  gm_always_assert (G_TRACEPTR >= (unsigned) gtracebufarea
		    && G_TRACEPTR <
		    (unsigned) (gtracebufarea) + sizeof (gtracebufarea));
#endif

#if 0
  gm_printf ("max_node_id = %d\n", gm.max_node_id);
#endif

  gm_always_assert ((char *) &gm.connection[gm.max_node_id] <=
		    (char *) GM_PORT);

  gm.initialized = 1;
}

/****************************************************************
 * Ethernet support functions
 ****************************************************************/

#if GM_EMULATE_BYTE_DMAS && (GM_DMA_GRANULARITY > 1)
/* inline function to ensure compiler detects logical shifts right, with
   fast unrolled loop implementation. */
static inline void
_gm_ether_byte_dma_emul_shift_shift (gm_u32_t * fw,	/* from word */
				     gm_u32_t * tw,	/* to word */
				     int ls,	/* left shift amount */
				     int rs,	/* right shift amount */
				     gm_u32_t rsm,	/* right shift mask */
				     int word_cnt)	/* words to write */
{
  gm_u32_t a, b;

  LOG_DISPATCH (0, "GM_EMULATE_BYTE_DMAS: " __FUNCTION__ " called\n");

  a = b = *fw++;
  switch (word_cnt & 0xf)
    {
    case 0x0:
      while (word_cnt > 0)
	{
	  b = *fw++;
	  *tw++ = (a << ls) | ((b >> rs) & rsm);
    case 0xf:
	  a = *fw++;
	  *tw++ = (b << ls) | ((a >> rs) & rsm);
    case 0xe:
	  b = *fw++;
	  *tw++ = (a << ls) | ((b >> rs) & rsm);
    case 0xd:
	  a = *fw++;
	  *tw++ = (b << ls) | ((a >> rs) & rsm);
    case 0xc:
	  b = *fw++;
	  *tw++ = (a << ls) | ((b >> rs) & rsm);
    case 0xb:
	  a = *fw++;
	  *tw++ = (b << ls) | ((a >> rs) & rsm);
    case 0xa:
	  b = *fw++;
	  *tw++ = (a << ls) | ((b >> rs) & rsm);
    case 0x9:
	  a = *fw++;
	  *tw++ = (b << ls) | ((a >> rs) & rsm);
    case 0x8:
	  b = *fw++;
	  *tw++ = (a << ls) | ((b >> rs) & rsm);
    case 0x7:
	  a = *fw++;
	  *tw++ = (b << ls) | ((a >> rs) & rsm);
    case 0x6:
	  b = *fw++;
	  *tw++ = (a << ls) | ((b >> rs) & rsm);
    case 0x5:
	  a = *fw++;
	  *tw++ = (b << ls) | ((a >> rs) & rsm);
    case 0x4:
	  b = *fw++;
	  *tw++ = (a << ls) | ((b >> rs) & rsm);
    case 0x3:
	  a = *fw++;
	  *tw++ = (b << ls) | ((a >> rs) & rsm);
    case 0x2:
	  b = *fw++;
	  *tw++ = (a << ls) | ((b >> rs) & rsm);
    case 0x1:
	  a = *fw++;
	  *tw++ = (b << ls) | ((a >> rs) & rsm);
	  word_cnt -= 0x10;
	}
    }
  LOG_DISPATCH (0, "GM_EMULATE_BYTE_DMAS: " __FUNCTION__ " called\n");
}

/* Special realignment function that shifts the LEN bytes at FROM to
   memory location (FROM+SHIFT).  The byte pointers FROM and
   (FROM+SHIFT) must be in the same word.

   NOTE: This function will corrupt any partword above or below the
   destination buffer, just like the LANai 4 DMA engine.  This is OK,
   since it is used to realign data that has just been DMAd by the
   LANai4 DMA engine, so the partwords are already corrupted.  Code
   before and after the DMA+realignment must save and restore these
   partwords.

   This function performs very efficient realignment, requiring about
   1 instruction per byte copied, by using only word reads and
   writes. */

static inline void
gm_ether_byte_dma_emul_shift (char *from, int shift, int len)
{
  gm_u32_t *tw;			/* from word */
  gm_u32_t *fw;			/* to word */
  int words_to_shift;

  LOG_DISPATCH (0, "GM_EMULATE_BYTE_DMAS: " __FUNCTION__ " called.\n");

  gm_assert (((long) from & ~3) == ((long) (from + shift) & ~3));
  fw = tw = (gm_u32_t *) ((long) from & ~3);
  if (shift > 0)
    fw--;

  words_to_shift = (len + 6) >> 2;	/* worst-case is (((3+len)+3)&~3)/4 */
  switch (shift)
    {
    case -3:
    case 1:
      _gm_ether_byte_dma_emul_shift_shift (fw, tw, 24, 8, 0xffffff,
					   words_to_shift);
      break;
    case -2:
    case 2:
      _gm_ether_byte_dma_emul_shift_shift (fw, tw, 16, 16, 0xffff,
					   words_to_shift);
      break;
    case -1:
    case 3:
      _gm_ether_byte_dma_emul_shift_shift (fw, tw, 8, 24, 0xff,
					   words_to_shift);
      break;
    case 0:
      return;
    default:
      gm_always_assert (0);
    }

  LOG_DISPATCH (0, "GM_EMULATE_BYTE_DMAS: " __FUNCTION__ " returning.\n");
}
#endif /* GM_EMULATE_BYTE_DMAS */

/****************************************************************
 * gm_dispatch ()
 ****************************************************************/

void
gm_dispatch (void)
{
  GM_HANDLER_CALLED ();
  
  SET_HANDLER (POLL_EVENT, idle,);
  SET_HANDLER (START_SDMA_EVENT, sdma__start_sdma_0,);
  SET_HANDLER (FINISH_SDMA_EVENT, sdma__finish_sdma,);
  SET_HANDLER (SEND_ACK_EVENT, send__start_sending_ack,);
  SET_HANDLER (START_SEND_EVENT, send__start_sending_chunk_0,);
  SET_HANDLER (FINISH_SEND_EVENT, send__finish_sending_chunk,);
  SET_HANDLER (FINISH_RECV_PACKET_EVENT, recv__got_chunk_0,);
  SET_HANDLER (START_RECV_PACKET_EVENT, recv__start_receiving_chunk_0,);
  SET_HANDLER (RECV_BUFFER_OVERFLOW_EVENT, recv__discard_overflow_0,);
  SET_HANDLER (START_RDMA_EVENT, rdma__rdma_chunk_0,);
  SET_HANDLER (FINISH_RDMA_EVENT, rdma__token_done,);
  SET_HANDLER (TIMER_EVENT, timer,);
  SET_HANDLER (DMA_INT_EVENT, dma__handle_dma_interrupt,);
  SET_HANDLER (SDMA_INT_EVENT, dma__handle_sdma_interrupt,);
  SET_HANDLER (RDMA_INT_EVENT, dma__handle_rdma_interrupt,);
  SET_HANDLER (DMA_PANIC_EVENT, dma__panic,);
  
  /* Fairness handlers */

  SET_HANDLER (FAIR_SDMA_RDMA_EVENT, fair__sdma_rdma_0,);
  SET_HANDLER (FAIR_SDMA_EVENT, fair__sdma_0,);
  SET_HANDLER (FAIR_RDMA_EVENT, fair__rdma_0,);

  /* handlers for (de)activating polling */

  gm.poll.idle_handler = GM_REFERENCE_LABEL (idle);
  gm.poll.active_handler = GM_REFERENCE_LABEL (sdma__poll_for_sdma_0);

  gm_assert (gm.free_send_chunk_cnt < 2 || NOTICED_NOT (SEND_PENDING));
  /* gm_print_state ( ); */
  /* gm_printf ("free_send_chunk_cnt: 0x%x\n", free_send_chunk_cnt); */

  DISPATCH (133, "first dispatch");
  gm_abort ();
}

#include "gm_mcp_reboot_node.h"

int				/* return "int" to prevent compiler warning */
main (void)
{
  gm_initialize ();
  gm_dispatch ();
  return 0;
}

#if 0
static void
gm_bootstrap_init (void)
{
  gm_initialize ();
}
#endif

/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  c-backslash-column:72
  tab-width:8
  End:
*/

