#! /bin/sh -c "echo This script should not be run directly.; exit 1"
################################################################
# Myricom GM networking software and documentation
# Copyright (c) 2001 by Myricom, Inc.
# All rights reserved.	 See the file `COPYING' for copyright notice.
################################################################
# author: glenn@myri.com
# $Source: /repository/gm/scripts/misc/gm_util.sh.in,v $
#
# This file implements generic utility functions shared by multiple
# GM scripts.

################################################################
# Automatic configuration
################################################################

gm_configured_prefix="/opt/gm"

# If possible, use the GM_PREFIX environmental variable as the prefix,
# else use the prefix specified on the configure command line when GM was
# built.  This allows GM_INSTALL to change the prefix for all other scripts
# via the environmental cover script created for
if test x"$GM_PREFIX" = x
then
    prefix="$gm_configured_prefix"
else
    prefix="$GM_PREFIX"
fi

target="i686-pc-linux-gnu"

################################################################
# Globals
################################################################

gm_this_script="$0"
# Set the HOST environmental variable if it is not set.
if test "x$HOSTNAME" = x
then 
    HOSTNAME=`hostname` || HOSTNAME=unknown
fi
export HOSTNAME

################################################################
# Utility Functions
################################################################

################
# Core printing support
################

# Print a dot to indicate progress, or a user-supplied message if
# verbose printing is activated.

gm_need_newline=0
gm_dots=""
gm_progress ()
{
    # If verbose printing is activated, print the filename instead
    # of a dot.
    
    if test $gm_verbose != 0
    then
	gm_echo "       " "$@"
	return 0
    fi

    # If we've printed too many dots on a line, start a new line.

    if test ".$gm_dots" = \
    "........................................................................."
    then
	echo
	gm_dots=""
    fi

    # if we're at the start of a new line, skip 8 chars.

    if test ".$gm_dots" = "."
    then
	printf  "        "
	gm_dots="........"
    fi

    # Print a dot and update our state.
    
    gm_dots=".${gm_dots}"
    printf .
    gm_need_newline=1
}

# like printf, only starting a new line if needed.
gm_printf ()
{
    test "$gm_need_newline" = 0 || echo
    printf "$@"
    gm_need_newline=1
}

# Like echo, only start a new line if needed.
gm_echo ()
{
    # Restore sane state if we have been doing gm_printf() or gm_progress().
    test "$gm_need_newline" = 0 || echo
    gm_need_newline=0
    gm_dots=""
    
    # The following is the standard autoconf hack to make "$@" behave when
    # there are no positional parameters
    echo ${1+"$@"}
}

# Like gm_echo, but only in verbose mode.

gm_verbose_echo ()
{
    test "$gm_verbose" = 0 || gm_echo "       " "$@"
}

################
# debugging
################

# if "$1" print the rest of the args, much like GM_PRINT in GM C code.

GM_PRINT ()
{
    flag=$1
    shift
    if test "$flag" -ne 0
    then
	gm_echo "$@"
    fi
}

# Print a very visable but not ugly warning

gm_warn ()
{
    # Convert all characters to stars
    stars=`echo $* | sed -e 's/./#/g'`
    
    # # Limit the number of stars to 72 to prevent ugly wrapping on screen
    # # for long prints on 80 character wide terminals.
    # star_limit="#########" # 9 stars
    # star_limit=`echo "$star_limit" | sed -e 's/#/########/g'` # 72 stars
    # stars=`echo "$stars" | sed -e 's@$\({star_limit}\).*@\1@'` # strip excess

    # Print the warning, bracketed by stars.
    gm_echo "${stars}"
    gm_echo "$@"
    gm_echo "${stars}"
}

# Print a warning and exit with an error

gm_die ()
{
    status=$?
    test status != 0 || status = 1
    gm_warn $*
    exit $status
}

# Execute a command, perhaps logging the output.

gm_verbose=0
gm_do ()
{
    test x`echo "$1"` = x`echo "$1" | sed -e 's@^gm_@@'` \
    || gm_die "Must not gm_do(\"${1}\") because it is not a native command."
    
    if test "$gm_verbose" != 0
    then
	gm_echo "        >" "$@"
	( "$@" )
    else
	# Currently, logging is not supported.

	log="/dev/null"
	echo "${USER}@${HOSTNAME}> $*" 2>&1 >> ${log}
	( "$@" 2>&1 ) >>${log}
    fi
}

# Return the directory component of the path to a file.

gm_dirname ()
{
    dirname=`echo "$1" | sed -e 's@/[^/]*$@@'` || return $?
    test "x$dirname" != "x$1" || dirname=.
    printf "%s" "$dirname"
}

# GM scripts assume 'pwd' returns a path with no symlinks.  This
# implementation ensures that this is true, even when using bash.

gm_pwd ()
{
    if test x"$BASH_VERSION" = x
    then
	pwd
    else
	pwd -P
    fi
}

# Return the unique "real" name of a directory, without links in the
# path.

gm_canonical_dirname ()
{
    dirname=`gm_dirname "$1"` || return $?
    cd "$dirname"  2>&1 > /dev/null && gm_pwd
}

# Return the basename of a path.  The basename is the path without
# the directory component.

gm_basename ()
{
    basename=`echo "%s" "$1" | sed -e 's@.*/@@'`
    printf "%s" "$basename"
}

# Echo the real file name of a possibly covered file.

gm_echo_real_file ()
{
    gm_real_file=`gm_dirname "$1"`"/."`gm_basename "$1"` \
    || gm_die "This should never happen."
    
    if test -f "$gm_real_file"
    then
	printf "%s" "$gm_real_file"
    else
	printf "%s" "$1"
    fi
}

# Take exactly 2 parameters and return 0 if the files are the same,
# nonzero otherwise.  In case of an error, return nonzero.

gm_same_file ()
{
    # Confirm this is called with exactly 2 parameters
    test $# -eq 2 || return $?

    # Compute the source directory
    fromdir=`gm_canonical_dirname "$1"` || return $?
    fromfile=`gm_basename "$1"` || return $?
    todir=`gm_canonical_dirname "$2"` || return $?
    tofile=`gm_basename "$2"` || return $?

    # Determine if the files are the same
    test "x${fromdir}/$fromfile" = "x${todir}/$tofile"
}

# Copy "$1" to "$2", as safely as we know how.  This means copy to a temp
# file in the same directory and then move to the ultimate destination for
# near atomic updates.

LN_S="ln -s"
gm_link_not_copy=0
gm_cp ()
{
    gm_cp_src="$1"
    gm_cp_dest="$2"

    # Confirm this is called with exactly 2 parameters
    test $# -eq 2 || return $?

    gm_verbose_echo "> gm_cp \"$gm_cp_src\" \"$gm_cp_dest\""
    
    # Return immediately if the 2 files being copied are the same file
    if gm_same_file "$gm_cp_src" "$gm_cp_dest"
    then
	return 0
    fi

    # Copy the real file instead of any cover file
    gm_cp_src=`gm_echo_real_file "$gm_cp_src"` \
    || gm_die This should never happen.
    
    # Remove the temporary file if it exists.  There is no need to catch
    # errors here, as they will be caught by the cp or ln below.
    test -f "$gm_cp_dest.${HOSTNAME}.$$" && rm -f "$gm_cp_dest.${HOSTNAME}.$$"

    # Copy file or create link
    if test $gm_link_not_copy = 0
    then
	cp "$gm_cp_src" "$gm_cp_dest.${HOSTNAME}.$$" || return $?
    else
	${LN_S} "$gm_cp_src" "$gm_cp_dest.${HOSTNAME}.$$" || return $?
    fi

    # Arrange for the file to be uninstalled by the uninstall script and
    # then move the file to the destination
    gm_log_uninstall_command "rm -f \"$gm_cp_dest\"" \
    && mv -f "$gm_cp_dest.${HOSTNAME}.$$" "$gm_cp_dest"
    status=$?
    if test $status != 0
    then
	rm -f "$gm_cp_dest.${HOSTNAME}.$$"
	return $status
    fi

    return $status
}

# Like gm_cp(), but always copy; never link.

gm_cp_never_link ()
{
    gm_cp_never_link_save="${gm_link_not_copy}"
    gm_link_not_copy=0
    gm_cp "$@"
    gm_link_not_copy="${gm_cp_never_link_save}"
}

# Make a directory and log its creation so it can be removed by the
# uninstall script.

gm_mkdir ()
{
    dirname="$1"

    test $# = 1 || gm_die "Too many parameters to gm_mkdir."

    gm_verbose_echo "> gm_mkdir \"$1\""

    # Don't bother to make existing directories
    test -d "$dirname" && return 0
    
    gm_log_uninstall_command "rmdir \"${dirname}\"" \
    && mkdir -p "$dirname"

    return $?
}

# Is the user running as root?

gm_running_as_root ()
{
    test x`whoami` = xroot
}

# Is the directory at $prefix properly configured by GM_INSTALL?

gm_make_id_matches ()
{
    dir="$1"
    test -f "${dir}/.gm_make_id" || return 1
    gm_saved_make_id="$GM_MAKE_ID"
    . "${dir}/.gm_make_id" || return 1
    test "$gm_saved_make_id" = "$GM_MAKE_ID" || return 1
    GM_MAKE_ID="$gm_saved_make_id"
    return 0
}

gm_installed_make_id_matches ()
{
    dir="$1"
    test -f "${dir}/etc/gm/.gm_installed_make_id" || return 1
    . "${dir}/etc/gm/.gm_installed_make_id" || return 1
    test "$GM_INSTALLED_MAKE_ID" = "$GM_MAKE_ID" || return 1
    return 0
}

################################################################
# Logging
################################################################

gm_start_install_log ()
{
    # Start logging, if possible, being careful to handle the case where the
    # log file cannot be written because the install is occuring in-place
    # in a non-writable directory.

    log_file="${prefix}/.gm_install_log_for_${HOSTNAME}"

    # Remove the log file since root can often do this even when the file
    # is not writable by root if the directory is writable.
    
    rm -f ${log_file} >/dev/null 2>&1
    
    # Create the log file using touch instead of ">" since using ">" can
    # result in a "Permission denied" message that is hard to supress.
    
    touch ${log_file} >/dev/null 2>&1
    
    # # Empty the file by overwriting it.  This is important because the "rm"
    # # above may fail if the file is writable but the directory is not.
    # # However, take care to use ">" only if the file is writable to prevent
    # # an error message.
    # 
    # test -w ${log_file} \
    # && ( gm_echo Starting install as `whoami` > ${log_file} )

    gm_install_log date
}
    
gm_install_log ()
{
    # Log only if the log file exists and is writable.
    
    test -w ${log_file} && ( echo "$@" >> ${log_file} 2>&1 )
}

################################################################
# Auto-uninstall support
################################################################

# This is a log of the commands needed to undo the install.  The
# install script appends the "uninstall" command to the log before
# each "install" command is executed to ensure that the log always
# includes all the commands needed to undo a partial install.

gm_uninstall_file=""

# Create an uninstall file, creating its directory if needed.  Make sure
# to put commands in the uninstall file to remove the uninstall file itself.

gm_create_uninstall_file ()
{
    # Remove any previous installation
    
    test -x "$1" \
    && gm_echo "Removing previous installation." \
    && (. "$1")

    gm_uninstall_file_dir=`gm_dirname "$1"`
    if test -d "$gm_uninstall_file_dir"
    then
        touch "$1" \
        && (echo "exit 0" > "$1") \
        && gm_uninstall_file="$1" \
	&& gm_do trap gm_execute_uninstall_log EXIT \
        && return 0

        status=$?
	gm_verbose_echo There was a problem.        
        rm -f "$1" 2>&1 > /dev/null
        return $status
    else
	gm_do mkdir -p "$gm_uninstall_file_dir" \
        && gm_do touch "$1" \
        && (echo "exit 0" > "$1") \
        && gm_uninstall_file="$1" \
	&& gm_do trap gm_execute_uninstall_log EXIT \
        && return 0
        
        status=$?
	gm_verbose_echo There was a problem.        
        gm_do rm -f "$1" 2>&1 > /dev/null
        gm_do rmdir "$gm_uninstall_file_dir" 2>&1 > /dev/null
        return $status
    fi
}

# like gm_log_uninstall_command, but don't assume the line is a command
# with output to supress.
__gm_log_uninstall_command ()
{
    # Make sure the uninstall file exists
    test x"$gm_uninstall_file" != x \
    || gm_die "Internal error: gm_uninstall_file is not set."
    
    # Prepend command to the start of the uninstall file.
    echo "$@" > "${gm_uninstall_file}.$$"
    cat "${gm_uninstall_file}" >> "${gm_uninstall_file}.$$" \
    && mv -f "${gm_uninstall_file}.$$" "${gm_uninstall_file}" \
    || rm -f "${gm_uninstall_file}.$$"
}

# Prepend command to the start of the uninstall file.
gm_log_uninstall_command ()
{
    __gm_log_uninstall_command "(" "$@" ")" "2>&1 > /dev/null"
}

# Close the uninstall log and make it executable.
gm_close_uninstall_log ()
{
    __gm_log_uninstall_command "#! /bin/sh" || return $?
    chmod +x "$gm_uninstall_file" || return $?
    trap EXIT
    gm_uninstall_file=""
}

gm_execute_uninstall_log ()
{
    gm_warn "Automatically cleaning up after an exception."
    
    test x"$gm_uninstall_file" != x \
    || gm_die "Internal error: gm_uninstall_file is not set"

    test -f "$gm_uninstall_file" \
    || gm_die "Internal error: \"$gm_uninstall_file\" is missing."
    
    . "$gm_uninstall_file" 2>&1 > /dev/null
}

################################################################
#
################################################################

# If we are on machine "$1", then do command "$2".  On all other machines, simply
# succeed.

gm_if_on ()
{
    test "$HOSTNAME" = "$1" || return 0
    shift
    "$@" || gm_warn Command \""$@"\" failed.
}

# # If we are on machine "$1", do command "$2"... and die if it fails.  On all
# # other machines, simply succeed.
# 
# gm_do_or_die_if_on ()
# {
#     gm_if_on "$@" || exit $?
# }

################################################################
# Libtool
################################################################

# wrap libtool in a subshell so we can redirect its output.  Don't
# call libtool directly, but only throught the wrappers below, which
# should handle any uninstall logging needed.

# Install a library or application using libtool.
gm_libtool_finished=0
gm_libtool_install ()
{
    # Make sure this was not called after gm_libtool_finish.
    # test ${gm_libtool_finished} = 0 \
    # || gm_die Internal error: gm_libtool_install called after gm_libtool_finish.

    gm_verbose_echo "$gm_libtool" "--mode=install" "$gm_arch_install_program" "$1" "$2"
    
    test $# -eq 2 \
    && rm -f "$2" \
    && gm_log_uninstall_command "rm -f \"$2\"" \
    && ( gm_do "$gm_libtool" "--mode=install" "$gm_arch_install_program" "$1" "$2" 2>&1 > /dev/null )
}

# Finish library installation.  This should be called only after any
# gm_libtool_install call.

gm_libtool_finish ()
{
    # No uninstallation required for this step command
    gm_do "$gm_libtool" --finish "$@"
    gm_libtool_finished=1
}

# Install a library

gm_install_lib ()
{
    gm_libtool_install "$@"
}

# Install an executable using libtool, but also create a cover script
# around it to set LD_LIBRARY_PATH and GM_PREFIX.

gm_install_program ()
{
    gm_install_program_src="$1"
    gm_install_program_dest="$2"

    if head -1 "${gm_install_program_src}" | grep "^#! */bin/sh" > /dev/null
    then
	gm_log_uninstall_command "rm -f \"${gm_install_program_dest}\"" \
	|| gm_die "Could not log uninstall command."

	( echo "#! /bin/sh" > "${gm_install_program_dest}" \
	&& echo "GM_PREFIX=\"${prefix}\"" >> "${gm_install_program_dest}" \
	&& cat "${gm_install_program_src}" >> "${gm_install_program_dest}" \
	) || gm_die "Error creating script \"${gm_install_program_dest}\""
	
	chmod a+x "${gm_install_program_dest}" \
	|| gm_die "Could not make \"${gm_install_program_dest}\" executable"
    else
	gm_libtool_install ${gm_install_program_src} ${gm_install_program_dest} \
	|| gm_die "Could not install ${gm_install_program_src} as ${gm_install_program_dest}."
    fi
}

# # Install a file that may be either an executable or a plain file.
# 
# gm_install_nonlib ()
# {
#     src="$1"
#     dest="$2"
#     if test -x "$src"
#     then
# 	  gm_install_program "$src" "$dest"
#     else
# 	  gm_cp "$src" "$dest"
#     fi
# }

gm_wait_for_driver_to_load ()
{
    gm_do "${prefix}/bin/gm_board_info" \
    || (sleep 1 ; \
	gm_do "${prefix}/bin/gm_board_info" \
	|| (sleep 2 ; \
	    gm_do "${prefix}/bin/gm_board_info" \
	    || (sleep 4 ; \
		gm_do "${prefix}/bin/gm_board_info" \
		|| (sleep 8 ; \
		    gm_do "${prefix}/bin/gm_board_info")))) \
    || gm_die "Driver failed to initialize any GM interfaces"
}
