/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997, 1998, 1999 by Myricom, Inc.                 *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/*
  gm_stress.c
  all-to-all test for GM. Run it first on host with lowest GM ID.
  finucane@myri.com
  gcc -g -o gm_stress gm_stress.c ~/gm/binary/lib/libgm.a -I ~/gm/include
*/

/*
To run it, type gm_stress on every host in a cluster. If you decide
to use options, they must be the same on all hosts. You can use a little rsh script
to launch it on a cluster:

racksaver1% /ufs/arpa/finucane/gm/tests# cat rsh_stress
#!/bin/bash

for i in `cat $1`
do
ssh $i $2 $3 $4 $5 $6 $7 $8 &
done
racksaver1% /ufs/arpa/finucane/gm/tests# 

racksaver1% /ufs/arpa/finucane/gm/tests# cat hosts
racksaver1
racksaver2
racksaver3
racksaver4
racksaver5
racksaver6
racksaver7

racksaver1% /ufs/arpa/finucane/gm/tests# ./rsh_stress hosts /ufs/finucane_home/gm/tests/gm_stress -f /ufs/finucane_home/gm/tests/hosts
racksaver1% /ufs/arpa/finucane/gm/tests# 7 hosts.
7 hosts.
7 hosts.
7 hosts.
7 hosts.
7 hosts.
7 hosts.

sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 93 / 92 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 93 / 90 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 93 / 93 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 88 / 89 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 88 / 91 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 88 / 86 MB/s
sent / received 1024 messages per host
in lengths up to 524240 bytes.
at a rate of 93 / 94 MB/s


It works like this: Every host sends a given number of random sized
packets out of a big chunk of gm_dma_malloc memory to each host
randomly and terminates when there is an error or when it has received
the given number of packets from each host, (including itself). It
uses gm_zone.c, which is part of libgm, for malloc and free from the
send heap.  If any host terminates without error then it means the
test has succeeded globally. Otherwise the host that detected the
error will terminate first, and then the others after sends fail. The
test prints out a host count before the test starts.

It sends to self and will work in loopback but not point-to-point,
where there is no route-to-self.

Because of congestion at the receiver, send timeouts can occur. If the
test fails with the "exceeded --max-resends" error on any node,
increase --max-resends.  Increasing receive buffering with
--min-receives is also an option.

The datagrams option adds some percentage of uncounted datagrams to
the stream. 100 percent is legal but will never terminate.

*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#define GM_STRONG_TYPES 0
#include "gm.h"

#define insist(e) do{if(!(e)){printf ("assertion failed. %s:%d (%s)\n", __FILE__,__LINE__, #e); goto exception;}}while (0)
#define pinsis(e, p) do{if(!(e)){printf p; printf ("\n"); goto exception;}}while (0)
#define gcheck(r, p) do{if(r != GM_SUCCESS){gm_perror (p, (enum gm_status) r); goto exception;}}while (0)
#define ZONE_FACTOR 4

#ifndef htonl
#define htonl gm_hton_u32
#endif

typedef struct message_t
{
  gm_u32_t crc;
  gm_u32_t sequence;
  gm_u32_t from;
  gm_u32_t to;
  gm_u32_t datagram;
  gm_u32_t length;
}
message_t;

typedef struct host_t
{
  int id;
  char hostname[GM_MAX_HOST_NAME_LEN + 1];
  gm_u64_t bytes_sent;
  gm_u64_t bytes_received;
  int dropping;
  unsigned messages_completed;
  unsigned messages_sent;
  unsigned messages_resent;
  unsigned messages_received;
  unsigned datagrams_sent;
  unsigned datagrams_received;
}
host_t;

static int failed;
static unsigned my_id;
static int my_index;
static int num_hosts;
static host_t *hosts;
static int send_tokens;
static int max_send_tokens;
static char *send_buffer;
static struct gm_zone *zone;
static unsigned long size_mask;
static unsigned biggest;
static int num_hostnames;
static char **hostnames;
static char *filename;
static int checksum;
static int datagram_percent;
static int unit = 0;
static int port_id = 2;
static int max_sends = 1024;
static int max_size = 19;
static unsigned max_length;
static int min_receives = 4;
static int alarmed = 0;
static gm_alarm_t alarm;
static int barrier_seconds = 60 * 2;
static int max_resends = 30;
static int same_size = 0;

static int initialized = 0;
static int opened = 0;
static int started = 0;

static void careful_exit (struct gm_port *port, int r)
{
  int i;

  if (hosts)
    gm_free (hosts);

  if (zone)
    gm_zone_destroy_zone (zone);

  if (hostnames)
    {
      for (i = 0; i < num_hostnames; i++)
	if (hostnames[i])
	  gm_free (hostnames[i]);
      gm_free (hostnames);
    }
  if (port && opened)
    gm_close (port);

  if (initialized)
    gm_finalize ();
  exit (r);
}

static void usage (char *option)
{
  if (option)
    printf ("unknown option %s\n", option);

  printf ("options:\n");

  printf ("-u, --unit=NUM            [0],      myrinet board number.\n");
  printf ("-p, --port=NUM            [2],      GM port number.\n");
  printf ("-n, --num-sends=NUM       [1024],   sends per host.\n");
  printf ("-m, --max-size=NUM        [19],     max send \"size\" (in log2 units).\n");
  printf ("-f, --file=FILENAME       [none],   file of hostnames, one per line\n");
  printf ("-c, --checksum            [off],    compute checksums.\n");
  printf ("-r, --min-receives=NUM    [4],      minimum number of receive buffers per size.\n");
  printf ("-b, --barrier-seconds=NUM [120],    timeout for barrier.\n");
  printf ("-x, --max-resends=NUM     [30],     max resends per host destination for send timeouts.\n");
  printf ("-s, --same-size           [off],    make all messages the same size (--max-size).\n");
  printf ("-d, --datagram=NUM        [0],      percent datagrams.\n");
  
  careful_exit (0, 1);
}

static int option (int i, int argc, char **argv, char *format, char *letter, int *v,
	int v_min, int v_max)
{
  int r = 0;
  insist (argv && format && letter && v);

  if ((r = sscanf (argv[i], format, v)) == 1 || (i + 1 < argc && !strcmp (argv[i], letter)))
  {
    if (!r)
      *v = atoi (argv[i + 1]);
    if (*v < v_min || *v > v_max)
    {
      printf ("%d is not between %d and %d\n", *v, v_min, v_max);
      usage (0);
    }
    return r ? 1 : 2;
  }
  exception:return 0;
}

static void sig_handler (int sig)
{
  failed = 1;
}

static int compare (const void *a, const void *b)
{
  return strcmp (*(char **) a, *(char **) b);
}

static int read_hostnames (char *filename, int num_hostnames, char **hostnames)
{
  FILE *fp;
  int count = 0;
  char *s, *e;
  char buffer[GM_MAX_HOST_NAME_LEN + 1];

  insist (filename);
  insist (num_hostnames >= 0);
  insist (!num_hostnames || hostnames);

  fp = fopen (filename, "r");
  pinsis (fp != 0, ("couldn't open \"%s\".", filename));

  while ((s = fgets (buffer, GM_MAX_HOST_NAME_LEN, fp)))
  {
    while (isspace (*s))
      s++;
    for (e = s + strlen (s) - 1; e >= s && isspace (*e); e--)
      *e = 0;
    if (*s)
    {
      if (num_hostnames && hostnames)
      {
	insist (count < num_hostnames);
	insist (hostnames[count]);
	strcpy (hostnames[count], s);
      }
      count++;
    }
  }

  pinsis (count, ("no hostnames in \"%s\".", filename));
  insist (!num_hostnames || count == num_hostnames);
  fclose (fp);

  return count;
  exception:return 0;
}

static int create_hostnames (char *filename)
{
  int i;

  insist (filename);

  if ((num_hostnames = read_hostnames (filename, 0, 0)) <= 0)
    return 0;

  hostnames = (char **) gm_calloc (num_hostnames, sizeof (char *));
  pinsis (hostnames, ("malloc failed for hostnames"));

  for (i = 0; i < num_hostnames; i++)
    {
      hostnames[i] = (char *) gm_malloc (GM_MAX_HOST_NAME_LEN);
      pinsis (hostnames[i], ("malloc failed for hostname %i", i));
    }

  i = read_hostnames (filename, num_hostnames, hostnames);
  insist (i == num_hostnames);
  qsort (hostnames, num_hostnames, sizeof (char *), compare);
  return 1;

  exception:
  if (hostnames)
    gm_free (hostnames);
  hostnames = 0;
  return 0;
}

static int get_hosts (struct gm_port *port, host_t ** hosts)
{
  unsigned max_id;
  int num_hosts, i, j, r;
  char a[6];
  char *s;

  insist (port && hosts);

  r = gm_max_node_id_inuse (port, &max_id);
  gcheck (r, "gm_max_node_id_inuse failed");
  insist (max_id > 0);

  num_hosts = 0;
  for (i = 0; i <= max_id; i++)
  {
    if (gm_node_id_to_unique_id (port, i, a) == GM_SUCCESS
	&& (a[0] || a[1] || a[2] || a[3] || a[4] || a[5]))
    {
      if ((s = gm_node_id_to_host_name (port, i))
	  && (!num_hostnames
	      || bsearch (&s, hostnames, num_hostnames, sizeof (char *),
			  compare)))
	num_hosts++;
    }
  }
  pinsis (num_hosts, ("no hosts"));
  *hosts = (host_t *) gm_calloc (num_hosts, sizeof (host_t));
  insist (*hosts);

  for (i = j = 0; i <= max_id; i++)
  {
    if (gm_node_id_to_unique_id (port, i, a) == GM_SUCCESS
	&& (a[0] || a[1] || a[2] || a[3] || a[4] || a[5]))
    {
      if ((s = gm_node_id_to_host_name (port, i))
	  && (!num_hostnames
	      || bsearch (&s, hostnames, num_hostnames, sizeof (char *),
			  compare)))
      {
	(*hosts)[j].id = i;
	strncpy ((*hosts)[j++].hostname, s, GM_MAX_HOST_NAME_LEN);
      }
    }
  }

  insist (j == num_hosts);
  pinsis (num_hosts, ("no hosts"));

  return num_hosts;

  exception:return 0;
}

static char*create_buffers (struct gm_port *port, unsigned max_length, int max_count, int min_receives)
{
  int r, i, j, k, n;
  char *p;

  insist (port);
  insist (min_receives > 0);
  
  for (i = gm_min_size_for_length (max_length), j = 1, k = 0;
       i >= gm_min_message_size (port); i--, k += j, j *= 2)
  {
    if (j > max_count)
      j = max_count;
    else if (j < min_receives)
      j = min_receives;
      
    insist (k + j <= gm_num_receive_tokens (port));

    p = (char *) gm_dma_malloc (port, gm_max_length_for_size (i) * j);
    pinsis (p, ("gm_dma_malloc failed"));
    insist (j > 0);

    for (n = 0; n < j; n++)
      gm_provide_receive_buffer_with_tag (port,
					  p +
					  n * gm_max_length_for_size (i), i,
					  GM_LOW_PRIORITY, 0);

    size_mask |= (1 << i);
  }
  insist (size_mask);

  r = gm_set_acceptable_sizes (port, GM_LOW_PRIORITY, size_mask);
  gcheck (r, "gm_set_acceptable_sizes failed");

  p = (char *) gm_dma_malloc (port, max_length * ZONE_FACTOR);
  pinsis (p, ("gm_dma_malloc failed"));

  return p;

  exception:return 0;
}

static void alarm_callback (void*c)
{
  alarmed = 1;
}

static int handle_packet (void *p, int length)
{
  message_t *m = (message_t *) p;
  int from, to, sequence, datagram;

  insist (p && hosts);
  insist (num_hosts > 0);
  insist (my_index >= 0 || my_index < num_hosts);

  pinsis (length >= sizeof (message_t), ("bad length"));
  pinsis (!checksum
	  || gm_crc ((char *) m + sizeof (unsigned),
		     length - sizeof (unsigned)) == htonl (m->crc),
	  ("bad crc"));

  to = htonl (m->to);
  from = htonl (m->from);
  sequence = htonl (m->sequence);
  datagram = htonl (m->datagram);

  pinsis (to == my_index, ("bad destination"));
  pinsis (from >= 0 && from < num_hosts, ("bad source"));

  if (!datagram)
  {
    pinsis (sequence == hosts[from].messages_received + 1,
	    ("bad sequence %d from %d, expected %d", sequence, from,
	     hosts[from].messages_received + 1));

    hosts[from].messages_received++;
  }
  else
    hosts[from].datagrams_received++;

  hosts[from].bytes_received += length;

  return 1;
  exception:return 0;
}

static int epilogue (time_t start_time)
{
  int i, resends = 0;
  gm_u64_t bytes_sent = 0;
  gm_u64_t bytes_received = 0;
  time_t stop_time = time (0);

  insist (hosts && num_hosts > 0 && my_index >= 0 && my_index < num_hosts);

  for (i = 0; i < num_hosts; i++)
  {
    insist (hosts[i].messages_sent == max_sends);
    insist (hosts[i].messages_completed == max_sends);
    insist (hosts[i].messages_received == max_sends);

    bytes_received += hosts[i].bytes_received;
    bytes_sent += hosts[i].bytes_sent;
    resends += hosts[i].messages_resent;
  }
  printf ("sent / received %d messages per host\n", max_sends);
  if (datagram_percent)
    printf ("not including datagrams\n");
  printf ("in lengths %s %d bytes.\n", same_size ? "of" : "up to", biggest);

  if (stop_time > start_time)
    printf ("at a rate of %u / %u MB/s\n",
	    (unsigned) (bytes_sent / (1024 * 1024) / (gm_u64_t) difftime (stop_time, start_time)),
	    (unsigned) (bytes_received / (1024 * 1024) / (gm_u64_t) difftime (stop_time, start_time)));

  if (resends)
    printf ("there were %d resends\n", resends);
  
  return 1;
  exception:return 0;
}

static void send_callback (struct gm_port *port, void *context, gm_status_t status);

static void drop_callback (struct gm_port *port, void *context, gm_status_t status)
{
  message_t *m = (message_t *) context;
  int length, to;

  insist (port && zone && context);

  to = htonl (m->to);
  length = htonl (m->length);
  
  insist (to >= 0 && to < num_hosts);
  insist (length >= sizeof (message_t) && length <= max_length + sizeof (message_t));
  
  insist (hosts[to].hostname && *hosts[to].hostname);
  insist (hosts [to].dropping);
  hosts [to].messages_resent++;
  
  gm_send_to_peer_with_callback (port, m, gm_min_size_for_length (length), length, GM_LOW_PRIORITY, hosts[to].id, send_callback, m);
  exception:;
}

void send_callback (struct gm_port *port, void *context, gm_status_t status)
{
  char s[GM_MAX_HOST_NAME_LEN * 2];
  message_t *m = (message_t *) context;
  int to;

  insist (port && zone && context);

  to = htonl (m->to);
  insist (to >= 0 && to < num_hosts);
  insist (hosts[to].hostname && *hosts[to].hostname);

  switch (status)
  {
    case GM_SEND_DROPPED:
      drop_callback (port, context, status);
      break;
    case GM_SUCCESS:
      gm_zone_free (zone, context);
      hosts [to].messages_completed++;
      insist (hosts [to].messages_completed <= hosts [to].messages_sent);
      if (hosts [to].messages_completed == hosts [to].messages_sent)
	hosts [to].dropping = 0;
      send_tokens++;
      break;
    case GM_SEND_TARGET_NODE_UNREACHABLE:
      failed = 1;
    case GM_SEND_TARGET_PORT_CLOSED:
      if (started)
	failed = 1;
      sprintf (s, "%s: send to %s failed", hosts [my_index].hostname, hosts[to].hostname);
      gm_perror (s, status);
    case GM_SEND_TIMED_OUT:
      if (!failed && hosts [to].messages_resent >= max_resends)
      {
	printf ("%s: send to %s failed: exceeded --max-resends=%d\n", hosts [my_index].hostname, hosts[to].hostname, max_resends);
	failed = 1;
      }
      hosts [to].dropping = 1;
      gm_drop_sends (port, GM_LOW_PRIORITY, hosts[to].id, port_id,
		     drop_callback, context);
      break;

    default:
      insist (0);
  }
  exception:;
}

static int receive (struct gm_port *port)
{
  gm_recv_event_t *e;

  insist (port);

  if (failed)
    return -1;

  e = gm_blocking_receive_no_spin (port);
  switch (gm_ntoh_u8 (e->recv.type))
  {
    case GM_RECV_EVENT:
      if (!handle_packet
	  (gm_ntohp (e->recv.buffer), gm_ntoh_u32 (e->recv.length)))
	return failed = -1;
      gm_provide_receive_buffer_with_tag (port, gm_ntohp (e->recv.buffer), gm_ntoh_u8 (e->recv.size), GM_LOW_PRIORITY, 0);
      return 1;
    case GM_ALARM_EVENT:
      alarmed = 0;
      gm_unknown (port, e);
      if (alarmed)
	return failed = -1;
    default:
      gm_unknown (port, e);
      break;
  }
  exception:return 0;
}

static int send (struct gm_port *port, int to, int num_bytes, int datagram)
{
  message_t *m;
  int size;
  int length;

  insist (port);
  insist (num_hosts > 0);
  insist (num_bytes >= 0);
  insist (to >= 0 && to < num_hosts);

  while (!(m = (message_t *) gm_zone_malloc (zone, sizeof (*m) + num_bytes))
	 && send_tokens < max_send_tokens)
    if (receive (port) < 0)
      return 0;

  insist (m && (char *) m >= send_buffer
	  && (char *) m < send_buffer + max_length * ZONE_FACTOR);
  insist (hosts[to].messages_sent < max_sends);

  if (datagram)
    hosts[to].datagrams_sent++;
  else
    hosts[to].messages_sent++;

  m->from = htonl (my_index);
  m->to = htonl (to);
  m->sequence = htonl (hosts[to].messages_sent);
  m->datagram = htonl (datagram);

  while (send_tokens == 0)
    if (receive (port) < 0)
      return 0;

  length = num_bytes + sizeof (*m);
  m->length = htonl (length);
  size = gm_min_size_for_length (length);

  hosts[to].bytes_sent += length;

  if (length > biggest)
    biggest = length;

  if (checksum)
    m->crc =
      htonl (gm_crc
	     ((char *) m + sizeof (unsigned), length - sizeof (unsigned)));

  if (size < gm_min_message_size (port))
    size = gm_min_message_size (port);

  insist (size_mask & (1 << size));

  send_tokens--;
  if (datagram)
  {
    insist (length <= gm_mtu (port));
    gm_datagram_send (port, m, size, length, GM_LOW_PRIORITY, hosts[to].id,
		      port_id, send_callback, m);
  }
  else
    gm_send_to_peer_with_callback (port, m, size, length, GM_LOW_PRIORITY,
				   hosts[to].id, send_callback, m);

  return 1;
  exception:return 0;
}


static int barrier (struct gm_port *port)
{
  int i, j;

  insist (port && num_hosts > 0 && hosts);
  insist (my_id > 0);
  insist (my_index >= 0);
  insist (send_buffer);
  insist (barrier_seconds > 0 && barrier_seconds <= 60 * 60);
  
  gm_set_alarm (port, &alarm, barrier_seconds * 1024 * 1024, alarm_callback, 0);
  alarmed = 0;
  
  if (my_id == hosts[0].id)
  {
    for (i = 0; i < num_hosts - 1;)
    {
      if (receive (port) > 0)
	i++;
      else if (alarmed)
      {
	printf ("barrier failed for:\n");
		
	for (j = 1; j < num_hosts; j++)
	{
	  if (!hosts [j].messages_received)
	    printf ("%s\n", hosts[j].hostname);
	}
	return 0;
      }
    }
    for (i = 1; i < num_hosts; i++)
      if (!send (port, i, 0, 0) || alarmed)
	return 0;
  }
  else
  {
    while (!failed && !alarmed && !hosts [0].messages_completed)
    {
      if (!send (port, 0, 0, 0))
	return 0;
      
      while (receive (port) <= 0 && !alarmed && !failed);
    }
  }

  if (!alarmed)
    gm_cancel_alarm (&alarm);
  
  return !failed && !alarmed;
  
  exception:return 0;
}

int get_random (struct gm_port *port, int *to, int *length, int *datagram)
{
  int i, h;

  insist (to && length && datagram);

  h = (rand () >> 13) % num_hosts;
  insist (h >= 0 && h < num_hosts);

  if (datagram_percent == 100)
    *datagram = 1;
  else if (datagram_percent == 0)
    *datagram = 0;
  else
    *datagram = rand () < RAND_MAX / 100 * datagram_percent;

  for (;;)
  {
    for (i = 0; i < num_hosts && !hosts[i].dropping; i++);
    if (i == num_hosts)
      break;
    while (hosts [i].dropping)
      if (receive (port) < 0)
	return 0;
  }
  
  for (i = 0; i < num_hosts && hosts[(h + i) % num_hosts].messages_sent == max_sends; i++);

  if (i == num_hosts)
  {
    for (;;)
    {
      for (i = 0; i < num_hosts &&  hosts[i].messages_received == max_sends; i++);
      if (i == num_hosts)
      {
	while (send_tokens != max_send_tokens && receive (port) >= 0);
	return 0;
      }
      if (receive (port) < 0)
	return 0;
    }
  }

  *to = (h + i) % num_hosts;
  insist (*to >= 0 && *to < num_hosts);
  insist (hosts[*to].messages_sent < max_sends);
  insist (!hosts[*to].dropping);

  insist (max_length >= sizeof (message_t));

  if (same_size)
    *length = max_length;
  else
    *length =
      max_length > sizeof (message_t) ?
      (rand () >> 9) %
      ((*datagram ? gm_mtu (port) : max_length) + 1 - sizeof (message_t)) : 0;

  insist (*length >= 0 && *length <= max_length);
  insist (!*datagram || *length <= gm_mtu (port) - sizeof (message_t));
  insist (size_mask & (1 << gm_min_size_for_length (*length)));

  return 1;
  exception:return 0;
}

#ifdef GM_BUILDING_GM
static int
gm_stress (int argc, char **argv)
#else
     int main  (int argc, char **argv)
#endif
{
  struct gm_port *port = 0;
  int i, r, to, size, datagram;
  void *m;
  time_t start_time;

  for (i = 1; i < argc; i++)
  {
    if (!strcmp (argv[i], "--help") || !strcmp (argv[i], "-h"))
      usage (0);

    if ((r = option (i, argc, argv, "--unit=%i", "-u", &unit, 0, 16))
	|| (r = option (i, argc, argv, "--port=%i", "-p", &port_id, 0, 16))
	|| (r = option (i, argc, argv, "--num-sends=%i", "-n", &max_sends, 1, 1024 * 1024 * 1024))
	|| (r = option (i, argc, argv, "--max-size=%i", "-m", &max_size, 4, 29))
	|| (r = option (i, argc, argv, "--min-receives=%i", "-r", &min_receives, 1, 1024))
	|| (r = option (i, argc, argv, "--barrier-seconds=%i", "-b", &barrier_seconds, 1, 60 * 60))
	|| (r = option (i, argc, argv, "--datagram=%i", "-d", &datagram_percent, 0, 100))
	|| (r = option (i, argc, argv, "--max-resends=%i", "-x", &max_resends, 0, 1024 * 1024 * 1024)))
    {
      i += r - 1;
      continue;
    }

    if (!strcmp (argv[i], "--checksum") || !strcmp (argv[i], "-c"))
      checksum = 1;
    else if (!strcmp (argv[i], "--same-size") || !strcmp (argv[i], "-s"))
      same_size = 1;
    else if (i < argc - 1 && !strcmp (argv[i], "-f"))
      filename = argv[++i];
    else if ((filename = strstr (argv[i], "--filename=")))
      filename += strlen ("--filename=");
    else
      usage (argv[i]);
  }

  if (filename && !create_hostnames (filename))
    careful_exit (port, 1);

  max_length = 2 << max_size;
  insist (max_length >= sizeof (message_t));

  initialized = gm_init ();
  gcheck (initialized, "gm_init failed");

  r =
    gm_open (&port, unit, port_id, "gm_stress",
	     (enum gm_api_version) GM_API_VERSION);
  gcheck (r, "gm_open failed");
  opened = 1;

  gm_initialize_alarm (&alarm);
  
  r = gm_get_node_id (port, &my_id);
  gcheck (r, "gm_get_node_id failed");

  if ((num_hosts = get_hosts (port, &hosts)) <= 0)
    careful_exit (port, 1);

  for (my_index = 0; my_index < num_hosts && hosts[my_index].id != my_id;
       my_index++);

  send_buffer = create_buffers (port, max_length, 10, min_receives);
  insist (send_buffer);

  zone = gm_zone_create_zone (send_buffer, max_length * ZONE_FACTOR);
  insist (zone);

  m = gm_zone_malloc (zone, max_length);
  insist (m);
  gm_zone_free (zone, m);

  max_send_tokens = send_tokens = gm_num_send_tokens (port) - 1;
  insist (send_tokens > 0);

  r = barrier (port);
  pinsis (r, ("barrier failed"));
  started = 1;
  
  srand (time (&start_time));
  signal (SIGINT, sig_handler);

  printf ("%d host%s.\n", num_hosts, num_hosts != 1 ? "s" : "");
  fflush (stdout);

  while (get_random (port, &to, &size, &datagram)
	 && send (port, to, size, datagram));

  if (!failed)
    epilogue (start_time);

  careful_exit (port, failed);

  exception:
  careful_exit (port, 1);
  return 1;
}

#ifdef GM_BUILDING_GM
#define GM_MAIN gm_stress
#include "gm_main.h"
#endif
