/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/* This fairly simple programs is used to automatically generate
   conservative Makefile dependency lists for the source files
   specified on stdin.  Unlike other similar programs, this program is
   designed to generate a single dependency file including dependencies
   for all architectures.

   The program works as follows: The user supplies on STDIN a list of
   all source files within the project, including all header files.
   The program scans these files for include lines, and then outputs a
   dependency list, indicating that each C source file depends on all
   header files with the same basename.  This conservative strategy
   works quite well for the GM source tree.

   This program has some GM-specific features.  It treats
   the mcp/, drivers/, and libgm/ directories specially. */


#include "gm_config.h"
#include "gm_call_trace.h"
#include "gm.h"

#ifdef STDC_HEADERS
#include <stdlib.h>		/* malloc */
#include <string.h>
#include <stdio.h>
#endif

#if defined HAVE_UNISTD_H
#include <unistd.h>		/* mmap, lseek */
#endif

#ifdef HAVE_MMAP
#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>		/* mmap */
#endif
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>		/* lseek */
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef WIN32
#include <io.h>
#define open _open
#define read _read
#define lseek _lseek
#define close _close
#define ssize_t int
#define getpagesize() 4096
#endif

/****************************************************************
 * Structures
 ****************************************************************/

/* Element in a list of files with a given basename. */

struct file_ref
{
  struct file_ref *next;
  struct file *file;
};

/* Element in a list of basenames of files included by a scanned file. */

struct basename_ref
{
  struct basename_ref *next;
  char *basename;
};

/* Element in a list of scanned files. */

struct file
{
  struct file *next;
  char *fullname;
  struct basename_ref *basename_refs;
};

/****************************************************************
 * globals
 ****************************************************************/

struct gm_hash *basename_to_file_refs;

/****************************************************************
 * mmap() emulation
 ****************************************************************/

#ifndef HAVE_MMAP

#ifndef MAP_PRIVATE
#define MAP_PRIVATE 1
#endif

#ifndef PROT_READ
#define PROT_READ 1
#endif

static void *
my_mmap (void *addr, size_t len, int prot, int flags, int fd, off_t off)
{
  off_t pos;
  char *buf;

  GM_CALLED ();
  
  /* save current position */

  pos = lseek (fd, 0, SEEK_CUR);
  if (pos == -1)
    {
      perror ("seek failed");
      goto abort_with_nothing;
    }

  /* seek to start of the region to mmap. */

  {
    off_t result;
    result = lseek (fd, off, SEEK_SET);
    if (result == -1)
      {
	perror ("seek failed");
	goto abort_with_nothing;
      }
  }

  /* allocate a buffer for the data */

  buf = (void *) gm_malloc (len + 1);
  if (!buf)
    {
      perror ("could not malloc for my_mmap()");
      goto abort_with_pos;
    }

  /* read the requested data into the buffer and null terminate */

  {
    ssize_t result;

    result = read (fd, buf, len);
    if (result == -1)
      {
	perror ("could not read in file.\n");
	goto abort_with_buf;
      }
    buf[len] = 0;
  }

  /* reset the file position. */

  pos = lseek (fd, pos, SEEK_SET);
  if (pos == -1)
    {
      perror ("could not reset pos in my_mmap()");
      goto abort_with_buf;
    }

  GM_RETURN (buf);

abort_with_buf:
  gm_free (buf);
abort_with_pos:
  lseek (fd, pos, SEEK_SET);
abort_with_nothing:
  GM_RETURN ((void *) -1);
}


int
my_munmap (void *ptr, size_t len)
{
  GM_CALLED ();
  gm_free (ptr);
  GM_RETURN (0);
}

#else /* HAVE_MMAP */

static void *
my_mmap (void *addr, size_t len, int prot, int flags, int fd, off_t off)
{
  GM_CALLED ();
  GM_RETURN (mmap (addr, len, prot, flags, fd, off));
}

static int
my_munmap (void *ptr, size_t len)
{
  GM_CALLED ();
  GM_RETURN (munmap (ptr, len));
}

#endif /* HAVE_MMAP */

/****************************************************************
 * Utility functions
 ****************************************************************/

static char *
gm_basename (char *fullname)
{
  char *b;

  GM_CALLED ();
  for (b = fullname + strlen (fullname) - 1; b > fullname; b--)
    {
      if (*b == '/')
	GM_RETURN (b + 1);
    }
  GM_RETURN (fullname);
}

static char *
suffix (char *fullname)
{
  char *b;

  GM_CALLED ();
  for (b = fullname + strlen (fullname) - 1; b > fullname; b--)
    {
      if (*b == '.')
	GM_RETURN (b);
    }
  GM_RETURN (fullname + strlen (fullname));	/* no suffix */
}

static char *
_rindex (const char *s, int c)
{
  const char *ret = 0;

  GM_CALLED ();
  while (*s)
    {
      if (*s == c)
	ret = s;
      s++;
    }
  GM_RETURN ((char *) ret);
}

/****************************************************************
 * Header output functions
 ****************************************************************/

static gm_status_t
output_headers1 (struct file *f, struct gm_hash *exclude)
{
  struct basename_ref *br;
  gm_status_t status;

  GM_CALLED ();
  if (gm_hash_find (exclude, f->fullname))
    GM_RETURN (GM_SUCCESS);

  fputs ("\\\n\t", stdout);
  fputs (f->fullname, stdout);
  fputs (suffix (f->fullname), stdout);
  
  status = gm_hash_insert (exclude, f->fullname, f);
  if (status != GM_SUCCESS)
    {
      fprintf (stderr, "Could not insert file in hash table.\n");
      GM_RETURN (status);
    }

  for (br = f->basename_refs; br; br = br->next)
    {
      struct file_ref *fr;

      for (fr = gm_hash_find (basename_to_file_refs, br->basename);
	   fr; fr = fr->next)
	{
	  status = output_headers1 (fr->file, exclude);
	  if (status != GM_SUCCESS)
	    GM_RETURN (status);
	}
    }

  GM_RETURN (GM_SUCCESS);
}

static gm_status_t
output_headers (struct file *f)
{
  struct basename_ref *br;
  struct gm_hash *exclude_hash;
  gm_status_t status;

  /* allocate resources, recursively output headers, and free resources */

  GM_CALLED ();
  exclude_hash
    = gm_create_hash (gm_hash_compare_strings, gm_hash_hash_string,
		      0, 0, 0, 0);
  if (!exclude_hash)
    {
      fprintf (stderr, "Could not create hash table.\n");
      GM_RETURN (GM_OUT_OF_MEMORY);
    }
  status = gm_hash_insert (exclude_hash, f->fullname, f);
  if (status != GM_SUCCESS)
    {
      goto abort_with_hash;
    }
  for (br = f->basename_refs; br; br = br->next)
    {
      struct file_ref *fr;

      for (fr = gm_hash_find (basename_to_file_refs, br->basename);
	   fr; fr = fr->next)
	{
	  status = output_headers1 (fr->file, exclude_hash);
	  if (status != GM_SUCCESS)
	    {
	      goto abort_with_hash;
	    }
	}
    }
  gm_destroy_hash (exclude_hash);

  GM_RETURN (GM_SUCCESS);

 abort_with_hash:
  gm_destroy_hash (exclude_hash);
  GM_RETURN (status);
}

/****************************************************************
 * Scanning
 ****************************************************************/

static gm_status_t
notice_file_basename (struct file *f)
{
  char *b;
  struct file_ref *fr, *first;
  gm_status_t status;

  GM_CALLED ();
  b = gm_basename (f->fullname);

  /* allocate storage for the file reference */

  fr = (struct file_ref *) gm_calloc (1, sizeof (struct file_ref));
  if (!fr)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  
  fr->file = f;

  /* add the file reference to the list of file references for the
     basename */

  first = gm_hash_find (basename_to_file_refs, b);
  if (first)
    {
      fr->next = first->next;
      first->next = fr;
    }
  else
    {
      status = gm_hash_insert (basename_to_file_refs, b, fr);
      if (status != GM_SUCCESS)
	goto abort_with_fr;
    }
  GM_RETURN (GM_SUCCESS);

abort_with_fr:
  gm_free (fr);
abort_with_nothing:
  GM_RETURN (status);
}

static
void
forget_file_basename (struct file *f)
{
  struct file_ref *fr, *next;
  
  fr = gm_hash_remove (basename_to_file_refs, gm_basename (f->fullname));
  while (fr)
    {
      next = fr->next;
      gm_free (fr);
      fr = next;
    }
}

/* record the fact that file F includes a file named BASENAME. */

static gm_status_t
record_basename_ref (struct file *f, char *basename)
{
  gm_status_t status;
  struct basename_ref *br;

  GM_CALLED ();
  br = (struct basename_ref *) gm_malloc (sizeof (struct basename_ref));
  if (!br)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_nothing;
    }
  br->basename = gm_strdup (basename);
  if (!br->basename)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_br;
    }
  br->next = f->basename_refs;
  f->basename_refs = br;
  GM_RETURN (GM_SUCCESS);

abort_with_br:
  gm_free (br);
abort_with_nothing:
  GM_RETURN (status);
}

/* Splice out a corresponding basename ref */
  
static void
forget_basename_ref (struct file *f, char *basename)
{
  struct basename_ref **br;

  for (br=&f->basename_refs; *br; br=&((*br)->next))
    {
      if (strcmp ((*br)->basename, basename) == 0)
	{
	  struct basename_ref *next;

	  next = (*br)->next;
	  gm_free ((*br)->basename);
	  gm_free (*br);
	  *br = next;
	  break;
	}
    }
}

/* Build a structure describing the file and all of the local includes
   that it references. */

static struct file *
scan (char *fullname)
{
  int fd;
  void *mm;
  char *cp, c, *tmp, *mark;
  struct file *f;
  long length;
  int err;

  /****************
   * Create the file record
   ****************/

  /* alloc */

  GM_CALLED ();
  f = (struct file *) gm_calloc (1, sizeof (struct file));
  if (!f)
    {
      perror ("could not malloc space for file");
      goto abort_with_nothing;
    }

  /* record the full relative path to the file */

  while (strncmp (fullname, "./", 2) == 0)
    fullname += 2;
  f->fullname = gm_strdup (fullname);
  if (!f->fullname)
    goto abort_with_f;

  /****************
   * record that basename maps to the file
   ****************/

  if (notice_file_basename (f) != GM_SUCCESS)
    goto abort_with_fullname;

  /****************
   * Scan the file for local include file references.
   ****************/

  fd = open (fullname, O_RDONLY);
  if (fd == -1)
    {
      perror ("could not open file");
      goto abort_with_fullname;
    }

  /* determine the length of the file. */

  length = lseek (fd, 0, SEEK_END);
  if (length == -1)
    {
      perror ("could not determine file length");
      goto abort_with_open;
    }

  /* round length up to a page boundary. */

  length = (length + getpagesize () - 1) & ~(getpagesize () - 1);
  if (!length)
    goto munmap_done;

  /* map the file. */

  mm = my_mmap (0, length, PROT_READ, MAP_PRIVATE, fd, 0);
  if (mm == 0 || mm == (void *) -1)
    {
      perror ("could not map file");
      goto abort_with_open;
    }

#define NEXT(cnt) do							\
  {									\
    cp += cnt;								\
    if ((char *) cp - (char *) mm >= length)				\
      goto scan_done;							\
    c = *cp;								\
    if (!c)								\
      goto scan_done;							\
  } while (0)

  cp = mm;
  c = *cp;
  while (1)
    {
      /* In this loop outside the NEXT() macro, cp points to the
         current character, and c is a copy of that character. */

      /* match "#"' */

      if (c != '#')
	goto next_line;
      NEXT (1);

      /* eat whitespace */

      while (c == ' ' || c == '\t')
	NEXT (1);

      /* match "include" */

      if (strncmp (cp, "include", 7) != 0)
	goto next_line;
      NEXT (7);

      /* match whitespace */

      if (c != ' ' && c != '\t')
	goto next_line;

      /* eat extra whitespace */

      while (c == ' ' || c == '\t')
	NEXT (1);

      /* match "\"" */

      if (c != '"' && c != '<')
	goto next_line;
      NEXT (1);

      /* eat and record filename.  (eat first to catch EOF) */

      mark = cp;
      while (c != '"' && c != '>')
	NEXT (1);
      tmp = (char *) gm_malloc (cp - mark + 1);
      if (!tmp)
	goto abort_with_my_mmap;
      gm_bcopy (mark, tmp, cp - mark);
      tmp[cp - mark] = 0;
      if (record_basename_ref (f, gm_basename (tmp)) != GM_SUCCESS)
	{
	  gm_free (tmp);
	  goto abort_with_my_mmap;
	}
      gm_free (tmp);

      /* find start of next line. */

    next_line:
      while (c != '\n')
	NEXT (1);
      NEXT (1);
    }
scan_done:

  err = my_munmap (mm, length);
  if (err)
    {
      perror ("could not munmap");
      goto abort_with_open;
    }
munmap_done:
  err = close (fd);
  if (err)
    {
      perror ("could not close");
      goto abort_with_fullname;
    }
  GM_RETURN (f);

abort_with_my_mmap:
  my_munmap (mm, length);
abort_with_open:
  close (fd);
abort_with_fullname:
  gm_free (f->fullname);
abort_with_f:
  gm_free (f);
abort_with_nothing:
  GM_RETURN (0);
}

static
void
unscan (struct file *f)
{
  /* free all the basename refs */

  while (f->basename_refs)
    {
      forget_basename_ref (f, f->basename_refs->basename);
    }
  forget_file_basename (f);
  gm_free (f->fullname);
  gm_free (f);
}

/****************************************************************
 * Extra files not included in the source release
 ****************************************************************/

char *generated_file[] = { "include/gm_auto_config.h", 0 };
struct file *generated_files;

static gm_status_t
notice_generated_file (char *fullname)
{
  gm_status_t status;
  struct file *f;

  GM_CALLED ();
  f = (struct file *) gm_calloc (1, sizeof (struct file));
  if (!f)
    {
      status = GM_OUT_OF_MEMORY;
      gm_perror ("could not malloc space for file", status);
      goto abort_with_nothing;
    }

  /* record the full relative_path to the file */

  while (strncmp (fullname, "./", 2) == 0)
    fullname += 2;
  f->fullname = gm_strdup (fullname);
  if (!f->fullname)
    {
      status = GM_OUT_OF_MEMORY;
      gm_perror ("could not strdup", status);
      goto abort_with_f;
    }

  /* notice the basename */

  status = notice_file_basename (f);
  if (status != GM_SUCCESS)
    {
      gm_perror ("could not notice file basename", status);
      goto abort_with_fullname;
    }

  f->next = generated_files;
  generated_files = f;
  GM_RETURN (GM_SUCCESS);

abort_with_fullname:
  gm_free (f->fullname);
abort_with_f:
  gm_free (f);
abort_with_nothing:
  GM_RETURN (status);
}

static
void
forget_generated_file (struct file *f)
{
  forget_file_basename (f);
  gm_free (f->fullname);
  gm_free (f);
}

static void
forget_generated_files (void)
{
  struct file *f;

  GM_CALLED ();
  while (generated_files)
    {
      f = generated_files;
      generated_files = f->next;
      forget_generated_file (f);
    }
  GM_RETURN_NOTHING ();
}

static gm_status_t
notice_generated_files (void)
{
  char **f = generated_file;
  gm_status_t status;
  
  GM_CALLED ();
  do
    {
      status = notice_generated_file (*f);
      if (status != GM_SUCCESS)
	{
	  forget_generated_files ();
	  GM_RETURN (status);
	}
    }
  while (*++f);

  GM_RETURN (GM_SUCCESS);
}

/****************************************************************
 * main program
 ****************************************************************/

int
main (int argc, char *argv[])
{
  struct file *file, *scanned_files = 0;
  char fullname[256];
  gm_status_t status = GM_SUCCESS;

  GM_CALLED ();
  status = gm_init ();
  if (status != GM_SUCCESS)
    GM_RETURN (1);

  /* initialize globals */

  basename_to_file_refs = gm_create_hash (gm_hash_compare_strings,
					  gm_hash_hash_string, 0, 0, 0, 0);
  if (!basename_to_file_refs)
    {
      status = GM_OUT_OF_MEMORY;
      fprintf (stderr,
	       "Could not create basename_to_file_refs hash table.\n");
      goto abort_with_hash;
    }

  /* scan each file and record its includes */

  while (fgets (fullname, 256, stdin))
    {
      int len;

      /* ignore empty lines */

      if (fullname[0] == 0 || fullname[0] == '\n')
	continue;

      /* strip trailing newline, if any */

      len = strlen (fullname);
      if (fullname[len - 1] == '\n')
	fullname[len - 1] = 0;

      /* scan the file for include files */

      file = scan (fullname);
      if (!file)
	{
	  status = GM_FAILURE;
	  goto abort_with_scanned_files;
	}
      file->next = scanned_files;
      scanned_files = file;
    }

  /* also notice the files that may need to be generated. */

  status = notice_generated_files ();
  if (status != GM_SUCCESS)
    {
      goto abort_with_scanned_files;
    }
  
  /* print the dependency list for each .c* file */

  for (file = scanned_files; file; file = file->next)
    {
      char *f, *s, save;

      f = file->fullname;
      s = suffix (f);

      /* output only .c* file dependencies */

      if (strncmp (s, ".c", 2) != 0)
	continue;

      /* create a timestamp to represent the dependencies by proxy.
	 This dramatically reduces the memory of Gnu make when it
	 reads the output. */

      fputs (f, stdout);
      fputs (suffix (f), stdout);
      fputs (":", stdout);

#if 0
      /* ouput dependency on the preprocessed version of the file. */

      fputs (f, stdout);
      fputs (suffix (f), stdout);
#endif
  
      /* output dependencies on the timestamps of all the included headers */

      status = output_headers (file);
      if (status != GM_SUCCESS)
	{
	  fprintf (stderr, "failed to output headers.\n");
	  goto abort_having_noticed_generated_files;
	}

      /* output a rule to touch the timestamp. */

      fputs ("\n", stdout);
    }

 abort_having_noticed_generated_files:
  forget_generated_files ();
 abort_with_scanned_files:
  while (scanned_files)
    {
      /* splice out a scanned file */
      
      file = scanned_files;
      scanned_files = file->next;

      unscan (file);
    }
 abort_with_hash:
  gm_destroy_hash (basename_to_file_refs);
  fflush (stdout);
  gm_exit (status);
  GM_RETURN (status);
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
